
#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_system.system.command.command import Command, CommandStates
from holado.common.context.session_context import SessionContext
from holado_system.system.filesystem.file import File
import cryptography.x509
from cryptography.hazmat.primitives import serialization
from holado.common.handlers.undefined import default_value

logger = logging.getLogger(__name__)


class KeyManager(object):
    """
    Helper for key management.
    """
    
    @classmethod
    def __get_path_manager(cls):
        return SessionContext.instance().path_manager
    
    @classmethod
    def generate_new_self_signed_key_files(cls, public_key_path, private_key_path, openssl_args):
        cls.__get_path_manager().makedirs(public_key_path)
        cls.__get_path_manager().makedirs(private_key_path)
        
        cmd = f"openssl req -out '{public_key_path}' -keyout '{private_key_path}' {openssl_args}"
        command = Command(cmd, do_log_output=True, do_raise_on_stderr=False, executable="/bin/bash")
        command.start()
        command.join()
        
        if command.state is not CommandStates.Success:
            raise TechnicalException(f"Error while executing openssl command [{cmd}] : [{command.stderr}]")
    
    @classmethod
    def generate_new_self_signed_key_files_for_localhost(cls, public_key_path, private_key_path, algorithm='rsa:2048'):
        KeyManager.generate_new_self_signed_key_files(
            public_key_path, private_key_path, 
            f"-x509 -newkey {algorithm} -noenc -sha256 -subj '/CN=localhost' -extensions EXT \
              -config <( printf \"[dn]\nCN=localhost\n[req]\ndistinguished_name = dn\n[EXT]\nsubjectAltName=DNS:localhost\nkeyUsage=digitalSignature\nextendedKeyUsage=serverAuth\")"
            )
    
    @classmethod
    def load_pem_public_key(cls, data=None, file_path=None, is_x509=False, **kwargs):
        if file_path is not None:
            data = File.read_file_content(file_path, mode='rb')
        
        if is_x509:
            cert = cryptography.x509.load_pem_x509_certificate(data, **kwargs)
            res = cert.public_key()
        else:
            res = serialization.load_pem_public_key(data, **kwargs)
        
        return res
    
    @classmethod
    def load_pem_private_key(cls, data=None, file_path=None, **kwargs):
        if file_path is not None:
            data = File.read_file_content(file_path, mode='rb')
        
        password = kwargs.get('password', None)
        res = serialization.load_pem_private_key(data, password, **kwargs)
        
        return res
    
    @classmethod
    def convert_public_key_to_hex(cls, public_key, encoding=default_value, public_format=default_value):
        """Convert public in a portable hexadecimal format.
        Note: default encoding is DER, and default public_format is SubjectPublicKeyInfo. If they are not compatible with public_key, appropriate values must be specified.
        """
        if encoding is default_value:
            encoding = serialization.Encoding.DER
        if public_format is default_value:
            public_format = serialization.PublicFormat.SubjectPublicKeyInfo
        return public_key.public_bytes(encoding=encoding, format=public_format).hex()
    
    
    





#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

from holado.common.context.session_context import SessionContext
import logging

logger = logging.getLogger(__name__)


class ScenarioStatusInfo(object):
    validation_category = None
    validation_status = None
    step_failed = None
    step_failed_nb = None
    scenario_context = None
    step_context = None
    scenario_error = None
    
    def __init__(self, category, status, step_failed, step_nb, scenario_context, step_context, scenario_error):
        self.validation_category = category
        self.validation_status = status
        self.step_failed = step_failed
        self.step_failed_nb = step_nb
        self.scenario_context = scenario_context
        self.step_context = step_context
        self.scenario_error = scenario_error


#TODO: make it a service in SessionContext
class ScenarioTools(object):
    
    _scenario_status_info_by_uid = {}
    
    @staticmethod
    def _get_scenario_context():
        return SessionContext.instance().get_scenario_context()

    @classmethod
    def _get_test_server_client(cls):
        return SessionContext.instance().test_server_client
    


    @classmethod
    def get_scenario_uid(cls, scenario):
        raise NotImplementedError()
    
    @classmethod
    def format_scenario_short_description(cls, scenario):
        raise NotImplementedError()
        
    #TODO: Remove the dependence to current scenario with _get_scenario_context, and find the scenario context in context history, 
    @classmethod
    def get_current_scenario_status_info(cls, scenario):
        return cls.get_scenario_status_info(scenario, cls._get_scenario_context())
        
    @classmethod
    def get_scenario_status_info(cls, scenario, scenario_context):
        scenario_uid = cls.get_scenario_uid(scenario)
        if scenario_uid not in cls._scenario_status_info_by_uid:
            info = cls._build_scenario_status_info(scenario, scenario_context)
            cls._scenario_status_info_by_uid[scenario_uid] = info
        return cls._scenario_status_info_by_uid[scenario_uid]
    
    @classmethod
    def _build_scenario_status_info(cls, scenario, scenario_context):
        step_failed, step_nb = cls.get_step_failed_info(scenario)
        
        # Define scenario status
        status = cls.compute_validation_status(scenario, scenario_context, step_failed)
        
        # Define scenario category
        category = cls.compute_validation_category(scenario, status)
        
        scenario_error = None
        if hasattr(scenario, "sut_error"):
            scenario_error = scenario.sut_error
        
        return ScenarioStatusInfo(category, status, step_failed, step_nb, 
                                  scenario_context,
                                  scenario_context.get_step(step_nb-1) if step_nb is not None else None,
                                  scenario_error)

    @classmethod
    def get_step_failed_info(cls, scenario):
        raise NotImplementedError()


    @classmethod
    def compute_validation_status(cls, scenario, scenario_context, step_failed):
        raise NotImplementedError()
        
    @classmethod
    def compute_validation_category(cls, scenario, status):
        if not cls._get_test_server_client().is_available:
            return None
        
        res = None
        
        # Get scenario execution statuses
        scenario_name = cls.format_scenario_short_description(scenario)
        sce_hist = cls._get_test_server_client().get_scenario_history(scenario_name=scenario_name, size=29)
        statuses = [s['status'] for s in reversed(sce_hist[0]['statuses'])] if sce_hist else []
        statuses.append(status)
        
        # Get scenario status sequences
        passed_sequences = []
        is_failed_relevant = None
        for status in statuses:
            if status == 'Passed':
                passed = True
            elif status.startswith("Failed"):
                passed = False
                if status == "Failed":
                    is_failed_relevant = True
                elif is_failed_relevant is None:
                    is_failed_relevant = False
            else:
                continue
            
            if len(passed_sequences) == 0 or passed != passed_sequences[-1][0]:
                passed_sequences.append([passed, 1])
            else:
                passed_sequences[-1][1] += 1
        
        # Compute category
        if passed_sequences:
            nb_exec = sum([x[1] for x in passed_sequences])
            last_passed, last_nb_times = passed_sequences[-1]
            if len(passed_sequences) == 1:
                if last_passed:
                    res = f'Always Success ({last_nb_times})'
                elif is_failed_relevant:
                    res = f'Always Failed ({last_nb_times})'
                else:
                    res = f'Always Not Relevant ({last_nb_times})'
            elif last_passed and len(passed_sequences) in [2, 3]:
                res = f'Fixed ({last_nb_times} success / {nb_exec})'
            elif len(passed_sequences) > 2:
                nb_fail = sum([x[1] for x in passed_sequences if not x[0]])
                if is_failed_relevant:
                    res = f'Random ({nb_fail} fails / {nb_exec})'
                else:
                    res = f'Random but Not Relevant ({nb_fail} fails / {nb_exec})'
            elif not last_passed:
                if last_nb_times == 1:
                    if is_failed_relevant:
                        res = f'Newly Failed ({last_nb_times} fails / {nb_exec})'
                    else:
                        res = f'Newly Failed but Not Relevant ({last_nb_times} fails / {nb_exec})'
                else:
                    if is_failed_relevant:
                        res = f'Regression ({last_nb_times} fails / {nb_exec})'
                    else:
                        res = f'Regression but Not Relevant ({last_nb_times} fails / {nb_exec})'
            else:
                res = f'Unknown (unmanaged sequence: {passed_sequences})'
        
        logger.debug(f"Category of scenario '{scenario}': {res}  (computed from last statuses: {statuses})")
        return res







#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from builtins import object

logger = None

def initialize_logger():
    global logger
    logger = logging.getLogger(__name__)


class ServiceManager(object):
    """
    Manage services in context types
    """
    
    def __init__(self, default_context_type):
        """Create service manager with given default context type.
        Default context type is usually SessionContext.
        """
        self.__default_context_type = default_context_type
        self.__services = {}
    
    @property
    def default_context_type(self):
        return self.__default_context_type
    
    def has_service(self, name):
        return name in self.__services
    
    def register_service_instance(self, name, service_instance, context, raise_if_service_exist=True, raise_if_object_exist=True):
        """
        Register a new service in a context with an existing instance.
        
        A property with service name is automatically added to context.
        
        Notes:
            It is possible to override an already registered service by passing "raise_if_service_exist=False, raise_if_object_exist=False".
            It is possible to replace an existing property by service property by passing "raise_if_service_exist=False".
        """
        from holado_core.common.exceptions.technical_exception import TechnicalException
        logger_ = logger if logger else logging
        
        logger_.trace(f"Registering service '{name}' in context {context}...")
        
        if raise_if_object_exist and context.has_object(name):
            raise TechnicalException(f"Context contains already an object '{name}'")
        if raise_if_service_exist and hasattr(context, name):
            raise TechnicalException(f"Context contains already an attribute '{name}'")
        
        # Register the service
        if context.has_object(name):
            context.remove_object(name)
        context.set_object(name, service_instance)
        
        # Add dedicated property in context
        self.add_context_property(name, type(context), raise_if_service_exist=False)
        
        logger_.debug(f"Registered service '{name}' instance {service_instance} in context {context}")
    
    def register_service_type(self, name, type_service, initialize_service=None, raise_if_exist=True, 
                              context_types=None, context_raise_if_service_exist=True, context_raise_if_object_exist=True, 
                              shortcut_in_types=None):
        """
        Register a new service type in service manager, and in defines context types.
        
        After this registration, it is possible to:
            - use the service by calling service manager property (the service instance is stored in service manager context)
            - register service in another context type with method register_service_type_in_context
        A property with service name is automatically added to context of given types.
        If shortcut_in_types is defined, a shortcut property is added in these types to access directly the context service
        (this feature currently works only for ProcessContext, ThreadContext, FeatureContext and ScenarioContext)
        
        When the context property is called the first time, the service is instantiated by calling "m = type_service()".
        After instantiation, it will be initialized by calling "initialize_service(m)".
        Initialization is typically used to inject dependencies, and process initialize actions needing these dependencies.
        
        Notes:
            If context_types=None, default context type is used
            Use lambda to define "type_service" when service "__init__" method has arguments.
            Use lambda to define "initialize_service" when service "initialize" method has arguments.
            It is possible to replace an existing attribute/property by service property by passing "raise_if_exist=False".
        """
        from holado_core.common.exceptions.technical_exception import TechnicalException
        logger_ = logger if logger else logging
        
        logger_.trace(f"Registering service '{name}'...")
        
        if shortcut_in_types and context_types and len(context_types) > 1:
            from holado_core.common.exceptions.functional_exception import FunctionalException
            raise FunctionalException(
                f"Impossible to define to which context type service the shortcuts must be linked to. \
                Please define only one context type when creating shortcuts")
        if raise_if_exist and name in self.__services:
            raise TechnicalException(f"A service '{name}' is already registered")
        if raise_if_exist and hasattr(self, name):
            raise TechnicalException(f"Service manager has already an attribute '{name}'")
        
        # Register the service
        self.__services[name] = (type_service, initialize_service, context_raise_if_service_exist, context_raise_if_object_exist)
        
        # Add dedicated property in service manager
        self.__add_service_manager_property(name)
        
        # Add property in context types
        _context_types = context_types
        if _context_types is None:
            _context_types = [self.__default_context_type]
        for context_type in _context_types:
            self.add_context_property(name, context_type, raise_if_exist)
        
        # Add shortcut property in given types
        if context_types and len(context_types) == 1:
            for type_ in shortcut_in_types:
                self.add_shortcut_property_to_context_service(name, type_, context_types[0], raise_if_exist)
        
        logger_.debug(f"Registered service '{name}'")
        # print(f"Registered service '{name}'")
    
    def register_service_type_in_context(self, name, context, raise_if_service_exist=None, raise_if_object_exist=None):
        logger_ = logger if logger else logging
        
        logger_.trace(f"Registering service '{name}' in context {context}...")
        
        self.__verify_object_doesnt_exist(name, context, raise_if_object_exist)
        
        if context.has_object(name):
            context.remove_object(name)
        self.add_context_property(name, type(context), raise_if_service_exist)
        
        logger_.trace(f"Registered service '{name}' in context {context}")
    
    def _get_context_service(self, name, context):
        """
        Get a service instance.
        If not already created, the instance is created according its registration and stored in context.
        """
        from holado_core.common.exceptions.technical_exception import TechnicalException
        logger_ = logger if logger else logging
        
        logger_.trace(f"Getting service '{name}' from {context}...")
        if not context.has_object(name):
            logger_.debug(f"Creating service '{name}'...")
            if name in self.__services:
                service = self.__services[name]
            else:
                raise TechnicalException(f"Unregistered service '{name}'")
            
            logger_.trace(f"Instantiating service '{name}'...")
            service_inst = service[0]()
            context.set_object(name, service_inst)
            logger_.debug(f"Service '{name}' is instantiated and stored in context {context}")
            
            if service[1] is not None:
                logger_.trace(f"Initializing service '{name}'...")
                service[1](service_inst)
            logger_.debug(f"Created service '{name}' for context {context}")
        return context.get_object(name)
    
    def add_context_property(self, name, context_type, raise_if_service_exist=None):
        # Note: next lines are commented, to allow to set again the service in context type
        # if raise_if_service_exist is None or raise_if_service_exist:
        #     self.__verify_service_doesnt_exist(name, context_type, raise_if_service_exist)
        if hasattr(context_type, name):
            logger.debug(f"Set again the property '{name}' in context type {context_type}")
        
        @property
        def context_service_property(self_context):
            return self._get_context_service(name, self_context)
        setattr(context_type, name, context_service_property)
        
        @property
        def context_has_service_property(self_context):
            return self_context.has_object(name)
        setattr(context_type, 'has_' + name, context_has_service_property)
    
    def add_shortcut_property_to_context_service(self, name, dst_type, context_type, raise_if_service_exist=None):
        from holado_multitask.multiprocessing.context.process_context import ProcessContext
        from holado.common.context.session_context import SessionContext
        from holado_multitask.multithreading.context.thread_context import ThreadContext
        from holado_test.common.context.feature_context import FeatureContext
        from holado_test.common.context.scenario_context import ScenarioContext
        from holado_core.common.exceptions.technical_exception import TechnicalException
        
        if issubclass(context_type, ProcessContext):
            self.__add_shortcut_property_to_context_service(name, dst_type, 
                                                            SessionContext.instance().multitask_manager.get_process_context,
                                                            raise_if_service_exist)
        elif issubclass(context_type, ThreadContext):
            self.__add_shortcut_property_to_context_service(name, dst_type, 
                                                            SessionContext.instance().multitask_manager.get_thread_context,
                                                            raise_if_service_exist)
        elif issubclass(context_type, FeatureContext):
            self.__add_shortcut_property_to_context_service(name, dst_type, 
                                                            SessionContext.instance().get_feature_context,
                                                            raise_if_service_exist)
        elif issubclass(context_type, ScenarioContext):
            self.__add_shortcut_property_to_context_service(name, dst_type, 
                                                            SessionContext.instance().get_scenario_context,
                                                            raise_if_service_exist)
        else:
            raise TechnicalException(f"Unmanaged context of type {context_type}")
    
    def __add_shortcut_property_to_context_service(self, name, dst_type, context_getter, raise_if_service_exist=None):
        from holado_core.common.exceptions.technical_exception import TechnicalException
        
        self.__verify_service_doesnt_exist(name, dst_type, raise_if_service_exist)
        
        @property
        def context_service_property(self_context):  # @UnusedVariable
            context = context_getter()
            if not hasattr(context, name):
                raise TechnicalException(f"Service '{name}' doesn't exist in context {context}")
            return getattr(context, name)
        setattr(dst_type, name, context_service_property)
    
    def __add_service_manager_property(self, name):
        @property
        def manager_service_property(self_manager):
            return self_manager._get_context_service(name, self_manager.context)
        setattr(self.__class__, name, manager_service_property)
    
    def __verify_service_exist(self, name):
        from holado_core.common.exceptions.technical_exception import TechnicalException
        
        if name not in self.__services:
            raise TechnicalException(f"Unregistered service '{name}'")
        
    def __verify_object_doesnt_exist(self, name, context, raise_if_object_exist=None):
        from holado_core.common.exceptions.technical_exception import TechnicalException
        
        if raise_if_object_exist is None:
            self.__verify_service_exist(name)
            raise_if_object_exist = self.__services[name][3]
        
        if raise_if_object_exist and context.has_object(name):
            raise TechnicalException(f"Context contains already an object '{name}'")
    
    def __verify_service_doesnt_exist(self, name, obj_inst_or_type, raise_if_service_exist=None):
        from holado_core.common.exceptions.technical_exception import TechnicalException
        
        if raise_if_service_exist is None:
            self.__verify_service_exist(name)
            raise_if_service_exist = self.__services[name][2]
            
        if raise_if_service_exist and hasattr(obj_inst_or_type, name):
            raise TechnicalException(f"Context contains already an attribute '{name}'")
        






#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

from holado_core.common.exceptions.functional_exception import FunctionalException
import logging
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado.common.handlers.object import DeleteableObject
from holado_core.common.tools.tools import Tools
from holado_docker.sdk.docker.container_logs import DockerContainerLogsFollower
import re
from holado_python.common.tools.datetime import DateTime
from holado.common.handlers.undefined import default_value
from holado_core.common.handlers.wait import WaitFuncResult
import time

logger = logging.getLogger(__name__)

try:
    import docker
    with_docker = True
except Exception as exc:
    if Tools.do_log(logger, logging.DEBUG):
        logger.debug(f"DockerClient is not available. Initialization failed on error: {exc}")
    with_docker = False


class DockerClient(object):
    @classmethod
    def is_available(cls):
        return with_docker
    
    def __init__(self):
        self.__client = docker.from_env()
        self.__containers = {}
        self.__volumes = {}
        
    @property
    def client(self):
        return self.__client
    
    def has_container(self, name, in_list=True, all_=False, reset_if_removed=True):
        # Note: Even if name exists in __containers, it is possible that the container has been removed
        if in_list:
            c = self.__get_container_from_list(name, all_=all_)
            res = c is not None
            
            if reset_if_removed and not res and name in self.__containers:
                del self.__containers[name]
        else:
            res = name in self.__containers
            
            if reset_if_removed and res:
                if self.__containers[name].status == "removed":
                    del self.__containers[name]
                res = False
        return res
    
    def get_container(self, name, all_=False, reset_if_removed=True):
        # Reset container if removed
        if reset_if_removed and name in self.__containers:
            if self.__containers[name].status == "removed":
                del self.__containers[name]
        
        # Get container from list if needed
        if name not in self.__containers:
            c = self.__get_container_from_list(name, all_=all_)
            if c:
                self.__containers[name] = DockerContainer(self, c)
        
        return self.__containers.get(name, None)
    
    def update_containers(self, all_=False, sparse=False, reset_if_removed=True):
        # Add new containers
        updated_names = set()
        for c in self.__client.containers.list(all=all_, sparse=sparse, ignore_removed=True):
            try:
                c_name = c.name
            except docker.errors.NotFound:
                # Container 'c' doesn't exist anymore
                continue
            
            if c_name not in self.__containers:
                self.__containers[c_name] = DockerContainer(self, c)
            updated_names.add(c_name)
            
        if reset_if_removed:
            for name in set(self.__containers.keys()).difference(updated_names):
                del self.__containers[name]
    
    def get_container_names(self, in_list=True, all_=False, sparse=False, include_patterns=None, exclude_patterns=None):
        incl_patterns = [ip if isinstance(ip, re.Pattern) else re.compile(ip) for ip in include_patterns] if include_patterns is not None else None
        excl_patterns = [ep if isinstance(ep, re.Pattern) else re.compile(ep) for ep in exclude_patterns] if exclude_patterns is not None else None

        if in_list:
            res = []
            for c in self.__client.containers.list(all=all_, sparse=sparse, ignore_removed=True):
                try:
                    c_name = c.name
                    if c_name is None:
                        c.reload()
                        c_name = c.name
                except docker.errors.NotFound:
                    # Container 'c' doesn't exist anymore
                    continue
                
                # Manage included and excluded patterns
                if excl_patterns is not None:
                    excluded = False
                    for ep in excl_patterns:
                        if ep.match(c_name):
                            excluded = True
                            break
                    if excluded:
                        continue
                if incl_patterns is not None:
                    included = False
                    for ip in incl_patterns:
                        if ip.match(c_name):
                            included = True
                            break
                    if not included:
                        continue
                
                res.append(c_name)
        else:
            res = list(self.__containers.keys())
        return res
    
    def __get_container_from_list(self, name, all_=False):
        res = None
        for c in self.__client.containers.list(all=all_, ignore_removed=True):
            try:
                c_name = c.name
            except docker.errors.NotFound:
                # Container 'c' doesn't exist anymore
                continue
            
            if c_name == name:
                res = c
                break
        return res
    
    def has_volume(self, name, in_list = False):
        res = name in self.__volumes
        if not res and in_list:
            v = self.__get_volume_from_list(name)
            res = v is not None
        return res
    
    def get_volume(self, name):
        if name not in self.__volumes:
            v = self.__get_volume_from_list(name)
            if v:
                self.__volumes[name] = DockerVolume(v)
        return self.__volumes.get(name)
    
    def get_all_volume_names(self):
        return [v.name for v in self.__client.volumes.list()]
    
    def __get_volume_from_list(self, name):
        res = None
        for v in self.__client.volumes.list():
            if v.name == name:
                res = v
                break
        return res
    
    def run_container(self, name, image, remove_existing=False, wait_started=True, auto_stop=True, **kwargs):
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"Running docker container '{name}' with image '{image}' and arguments {kwargs}{', and waiting it is started' if wait_started else ''}")
        
        # Manage remove if already existing
        cont = self.get_container(name)
        if cont:
            if remove_existing:
                if cont.status == "running":
                    if Tools.do_log(logger, logging.DEBUG):
                        logger.debug(f"Docker container '{name}' is running, stopping it before remove")
                    self.stop_container(name)
                
                if self.has_container(name):    # After stop, container is able to be automatically removed depending on its run parameters 
                    self.remove_container(name)
            else:
                logger.info(f"Docker container '{name}' is already running")
                return
        
        # Run container
        c = self.__client.containers.run(image, name=name, detach=True, **kwargs)
        container = DockerContainer(self, c)
        self.__containers[name] = container
        
        # Manage wait running status
        if wait_started:
            try:
                container.await_started()
            except:
                raise TechnicalException(f"Failed to run container of name '{name}' (status: {container.status})")
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"Run docker container '{name}' with image '{image}' and arguments {kwargs}{', and waited it is started' if wait_started else ''}")
            
        # Set properties
        container.auto_stop = auto_stop
        
        return container
    
    def restart_container(self, name, wait_started=True, timeout=default_value, **kwargs):
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"Restarting docker container '{name}' with arguments {kwargs}{', and waiting it is started' if wait_started else ''}")
        container = self.get_container(name)
        if not container:
            raise FunctionalException("Container of name '{}' doesn't exist")
        
        container.restart(**kwargs)
        
        if wait_started:
            try:
                container.await_started(timeout=timeout)
            except:
                raise TechnicalException(f"Failed to restart container of name '{name}' (status: {container.status})")
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"Restarted docker container '{name}' with arguments {kwargs}{', and waited it is started' if wait_started else ''}")
    
    def start_container(self, name, wait_started=True, timeout=default_value, **kwargs):
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"Starting docker container '{name}' with arguments {kwargs}{', and waiting it is started' if wait_started else ''}")
        container = self.get_container(name, all_=True)
        if not container:
            raise FunctionalException("Container of name '{}' doesn't exist")
        
        container.start(**kwargs)
        
        if wait_started:
            try:
                container.await_started(timeout=timeout)
            except:
                raise TechnicalException(f"Failed to start container of name '{name}' (status: {container.status})")
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"Started docker container '{name}' with arguments {kwargs}{', and waited it is started' if wait_started else ''}")
        
    def stop_container(self, name, wait_stopped=True, timeout=default_value, stop_kwargs=None, wait_kwargs=None):
        stop_kwargs = stop_kwargs if stop_kwargs is not None else {}
        wait_kwargs = wait_kwargs if wait_kwargs is not None else {}
        if timeout is not default_value:
            wait_kwargs['timeout'] = timeout
        
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"Stopping docker container of name '{name}'{', and waiting it is stopped' if wait_stopped else ''}")
        if name not in self.__containers:
            raise FunctionalException("Unknown container of name '{}'".format(name))
        elif self.__containers[name].status != "running":
            raise FunctionalException("Container of name '{}' is not running (status: {})".format(name, self.__containers[name].status))
        
        self.__containers[name].stop(**stop_kwargs)
        if wait_stopped:
            try:
                self.__containers[name].wait(**wait_kwargs)
            except docker.errors.NotFound:
                # This exception occurs on containers automatically removed on stop
                pass
        
        if wait_stopped and self.__containers[name].status == "running":
            raise FunctionalException("Failed to stop container of name '{}' (status: {})".format(name, self.__containers[name].status))
        else:
            if Tools.do_log(logger, logging.DEBUG):
                logger.debug(f"Stopped docker container of name '{name}'{', and waited it is stopped' if wait_stopped else ''}")
        
    def remove_container(self, name):
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"Removing docker container of name '{name}'")
        if not self.has_container(name, in_list=True, all_=True):
            raise FunctionalException(f"Container of name '{name}' doesn't exist")
        
        if name in self.__containers:
            del self.__containers[name]
        self.client.api.remove_container(name)
        
        if self.has_container(name, in_list=True, all_=True):
            raise FunctionalException(f"Failed to remove container of name '{name}'")
        else:
            if Tools.do_log(logger, logging.DEBUG):
                logger.debug(f"Removed docker container of name '{name}'")
    
    def await_container_exists(self, name, timeout=default_value):
        wait_status = WaitFuncResult(f"wait container '{name}' exists", 
                                      lambda: self.has_container(name, in_list=True, all_=True, reset_if_removed=False) )
        wait_status.redo_until(True)
        if timeout is not default_value:
            wait_status.with_timeout(timeout)
        return wait_status.execute()
    
        
class DockerContainer(DeleteableObject):
    def __init__(self, docker_client, container):
        super().__init__(container.name)
        
        self.__docker_client = docker_client
        self.__container = container
        self.__auto_stop = False
        
        # Manage logs
        self.__logs_follower = None
        self.__logs_formatter = None
        
    def _delete_object(self):
        if self.auto_stop and self.status == "running" and self.__docker_client and self.__docker_client.has_container(self.name):
            self.__docker_client.stop_container(self.name)
        
    @property
    def container(self):
        return self.__container
        
    @property
    def information(self):
        self.reload(raise_if_removed=False)     # If container doesn't exist anymore, use last known information
        return self.__container.attrs

    @property
    def status(self):
        """ Container status (created, restarting, running, removing, paused, exited, dead, or removed)
        """
        if not self.reload(raise_if_removed=False):
            return "removed"
        return self.__container.status

    @property
    def exit_code(self):
        """ Container status (created, restarting, running, removing, paused, exited, dead, or removed)
        """
        self.reload(raise_if_removed=False)     # If container doesn't exist anymore, use last known information
        return self.__container.attrs.get('State', {}).get('ExitCode', None)

    @property
    def health_status(self):
        """ Container health status (starting, healthy, unhealthy, or removed).
        """
        if not self.reload(raise_if_removed=False):
            return "removed"

        res = self.__container.health
        if res == 'unknown':
            return None
        else:
            return res
        
    @property
    def auto_stop(self):
        return self.__auto_stop
        
    @auto_stop.setter
    def auto_stop(self, auto_stop):
        self.__auto_stop = auto_stop
    
    def reload(self, raise_if_removed=False):
        """ Reload the container
        If raise_if_removed is False, returns if reload is in success (ie returns False if the container is removed).
        """
        res = True
        try:
            self.__container.reload()
        except docker.errors.NotFound:
            if raise_if_removed:
                raise
            else:
                res = False
        return res
        
    def restart(self, **kwargs):
        return self.__container.restart(**kwargs)
    
    def start(self, **kwargs):
        return self.__container.start(**kwargs)
    
    def stop(self, **kwargs):
        return self.__container.stop(**kwargs)
    
    def wait(self, **kwargs):
        return self.__container.wait(**kwargs)
    
    def await_started(self, timeout=default_value):
        health_status = self.health_status
        if health_status == "healthy":
            return
        elif health_status not in [None, 'removed']:
            return self.await_health_status("healthy", timeout=timeout)
        else:
            # Container hasn't healthy mechanism
            return self.await_status("running", timeout=timeout)
            #TODO: for containers that haven't healthy mechanism, add a mechanism to customize how health is computed
    
    def await_status(self, status, timeout=default_value):
        wait_status = WaitFuncResult(f"wait container '{self.name}' has status '{status}'", 
                                      lambda: self.status)
        wait_status.with_process_in_thread(False) \
                   .redo_until(status)
        if timeout is not default_value:
            wait_status.with_timeout(timeout)
        return wait_status.execute()
    
    def await_health_status(self, status, timeout=default_value):
        wait_status = WaitFuncResult(f"wait container '{self.name}' has health status '{status}'", 
                                      lambda: self.health_status)
        wait_status.with_process_in_thread(False) \
                   .redo_until(status)
        if timeout is not default_value:
            wait_status.with_timeout(timeout)
        return wait_status.execute()
    
    
    
    # Manage logs
    
    @property
    def is_following_logs(self):
        return self.__logs_follower is not None
    
    @property
    def logs_formatter(self):
        return self.__logs_formatter
    
    @logs_formatter.setter
    def logs_formatter(self, formatter):
        self.__logs_formatter = formatter
    
    def get_container_logs(self, stdout=True, stderr=True, timestamps=False, tail='all', since=None, until=None, wait_before_s=0.1):
        """ Get logs from Docker Engine
        Note: There is a latency between the moment a container makes a log and the moment it is available in Docker Engine.
              Thus this method waits by default a little time (wait_before_s) before getting logs.
              This feature can be neutralized by passing wait_before_s=None.
        @param stdout: If True, get stdout logs (default: True)
        @param stderr: If True, get stderr logs (default: True)
        @param timestamps: If True, add timestamps in logs (default: False)
        @param tail: Get specified number of lines at the end of logs. Either an integer of number of lines or the string 'all' (default: 'all')
        @param since: Since datetime (default: None)
        @param until: Until datetime (default: None)
        @param wait_before_s: None or number of seconds to wait before getting logs (default: 0.1)
        """
        if wait_before_s is not None:
            time.sleep(wait_before_s)
        
        if isinstance(since, str):
            since = DateTime.str_2_datetime(since)
        if isinstance(until, str):
            until = DateTime.str_2_datetime(until)
            
        logs_bytes = self.__container.logs(stdout=stdout, stderr=stderr, timestamps=timestamps, tail=tail, since=since, until=until)
        return self._get_logs_from_logs_bytes(logs_bytes, with_timestamps=timestamps)
    
    def _get_logs_from_logs_bytes(self, logs_bytes, with_timestamps=False):
        log_bytes_list = logs_bytes.split(b'\n')
        
        res = []
        for log_bytes in log_bytes_list:
            if not log_bytes:
                continue
            res.append(self._get_log_from_log_bytes(log_bytes, with_timestamps))
        
        return res
    
    def _get_log_from_log_bytes(self, log_bytes, with_timestamps=False):
        if with_timestamps:
            t_bytes, log_bytes = log_bytes.split(b' ', maxsplit=1)
            log_time = t_bytes.decode()
        
        try:
            log = log_bytes.decode()
        except:
            log = log_bytes
        
        if with_timestamps:
            res = (log_time, log)
        else:
            res = log
        
        return res
    
    def follow_logs(self, stdout=True, stderr=True, tail='all', since=None):
        """Follow logs from Docker Engine
        """
        if self.__logs_follower is not None:
            raise TechnicalException(f"Logs of container '{self.name}' are already followed")
        
        self.__logs_follower = DockerContainerLogsFollower(self)
        self.__logs_follower.follow_logs(stdout, stderr, tail, since)

    def reset_logs(self):
        """Reset followed logs
        """
        if self.__logs_follower is None:
            raise TechnicalException(f"Logs of container '{self.name}' are not followed")
        
        self.__logs_follower.reset()

    def get_logs(self, stdout=True, stderr=True, timestamps=False, tail='all', since=None, until=None, wait_before_s=0.1, formatter=None):
        if self.is_following_logs:
            logs = self.__logs_follower.get_logs(stdout=stdout, stderr=stderr, timestamps=timestamps, tail=tail, since=since, until=until, wait_before_s=wait_before_s)
        else:
            logs = self.get_container_logs(stdout=stdout, stderr=stderr, timestamps=timestamps, tail=tail, since=since, until=until, wait_before_s=wait_before_s)
        
        formatter = formatter if formatter is not None else self.__logs_formatter
        if formatter is not None:
            res = formatter.format_logs(logs, with_timestamps=timestamps)
        else:
            res = logs
        
        return res




class DockerVolume(object):
    def __init__(self, volume):
        self.__volume = volume
        
    @property
    def volume(self):
        self.__volume
        
    @property
    def attributes(self):
        self.__volume.reload()
        return self.__volume.attrs
    
    def get_attribute(self, attr_path):
        names = attr_path.split('.')
        attrs = self.attributes
        res = attrs
        for i, name in enumerate(names):
            if name in res:
                res = res[name]
            else:
                raise FunctionalException(f"Attribute '{'.'.join(names[:i+1])}' doesn't exist (requested attribute: '{attr_path}' ; volume attributes: {attrs})")
        return res
    




#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import pprint
import sys
import threading
import subprocess
import logging
import time
from enum import IntEnum
from holado_core.common.exceptions.functional_exception import FunctionalException
import copy
import signal
from holado_core.common.tools.tools import Tools
from holado_multitask.multitasking.multitask_manager import MultitaskManager

logger = logging.getLogger(__name__)


class CommandStates(IntEnum):
    """States of a command."""
    Ready = 1
    Running = 2
    Success = 3
    Error = 4


class Command(threading.Thread):
    """
    Execute a command in a thread.
    """

    def __init__(self, cmd, do_log_output = False, do_raise_on_stderr = False, **subprocess_kwargs):
        """
        'cmd' argument can be a list or a str. If the global command should contain a '"' in any argument, it is recommanded to pass 'cmd' as a str.
        """
        super().__init__()

        # Put thread as daemon
        self.daemon = True
        
        self.cmd = cmd
        self.__do_log_output = do_log_output
        self.__do_raise_on_stderr = do_raise_on_stderr
        
        self.__process = None
        self.__state = CommandStates.Ready
        self.__stdout = None
        self.__stderr = None
        self.__error = None
        self.__callback = None
        self.__callback_delay_ms = None
        self.__external_parameters = {}
        self.__subprocess_kwargs = subprocess_kwargs
        self.__stop_signal = signal.SIGTERM

 
    @property
    def stdout(self):
        return self.__stdout
 
    @property
    def stderr(self):
        return self.__stderr
 
    @property
    def error(self):
        return self.__error
 
    @error.setter
    def error(self, err):
        """Set error (an exception)."""
        if not isinstance(err, Exception):
            raise Exception("An exception is expected")
        self.__error = err
        self.__state = CommandStates.Error

    @property
    def _internal_process(self):
        return self.__process
    
    @property
    def state(self):
        return self.__state
    
    @property
    def return_code(self):
        if self.__process:
            return self.__process.returncode
        else:
            return None
    
    @property
    def callback(self):
        return self.__callback
    
    @callback.setter
    def callback(self, callback):
        """Set callback called when execution end."""
        self.__callback = callback
    
    @property
    def callback_delay_ms(self):
        return self.__callback_delay_ms
    
    @callback_delay_ms.setter
    def callback_delay_ms(self, delay_ms):
        """Set callback delay in ms."""
        self.__callback_delay_ms = delay_ms
    
    @property
    def external_parameters(self):
        return self.__external_parameters
    
    @property
    def stop_signal(self):
        return self.__stop_signal
    
    @stop_signal.setter
    def stop_signal(self, stop_signal):
        """Set stop signal."""
        self.__stop_signal = stop_signal
    
    def run(self):
        logger.debug("Call command: {}".format(self.cmd))
        try:
            self.__state = CommandStates.Running
            
            kwargs = copy.copy(self.__subprocess_kwargs)
            kwargs['universal_newlines'] = True
            if isinstance(self.cmd, str):
                self.__process = subprocess.Popen(self.cmd,
                                                  stdout=subprocess.PIPE,
                                                  stderr=subprocess.PIPE,
                                                  shell=True,
                                                  **kwargs)
            else:
                self.__process = subprocess.Popen(self.cmd,
                                                  stdout=subprocess.PIPE,
                                                  stderr=subprocess.PIPE,
                                                  shell=False,
                                                  **kwargs)
            
            self.__stdout = ""
            self.__stderr = ""
            while self.__process.returncode is None:
                # Wait a small time and in same time yield this thread
                time.sleep(0.001)
                
                self.__process.poll()
                self.__read_stdout_stderr()
            
            # Manage end status
            if self.return_code == 0:
                self.__state = CommandStates.Success
            else:
                self.__state = CommandStates.Error
            
            # Manage error
            if self.__do_raise_on_stderr:
                self.__raise_exception_if_stderr()

            # Print result
            if self.__do_log_output:
                msg = "Output of command {}:\n{}\n{}\n{}".format(self.cmd, '<'*10, self.stdout, '>'*10)
                logger.debug(msg)

        except Exception as exc:
            self.error = exc
            
        finally:
            if self.callback:
                if self.callback_delay_ms:
                    delay_s = self.callback_delay_ms / 1000.
                    logger.debug(f"Command [{self.cmd}] has finished, call callback in {delay_s} seconds")
                    t = threading.Timer(delay_s, self.callback, [self])
                    t.start()
                else:
                    logger.debug(f"Command [{self.cmd}] has finished, calling callback...")
                    self.callback(self)
            elif self.error is not None:
                logger.error(f"Command [{self.cmd}] has finished on error: {self.error}")
            elif self.state == CommandStates.Error:
                logger.error(f"Command [{self.cmd}] has finished on error code {self.return_code} and stderr: {self.stderr}")
            elif self.state != CommandStates.Success:
                logger.warning(f"Command [{self.cmd}] has finished with status {self.state.name}")
            else:
                logger.debug(f"Command [{self.cmd}] has succeeded")
    
    def __read_stdout_stderr(self):
        # Get results
        out = self.__process.stdout.read()
        if len(out) > 0:
            if sys.version_info < (3,0):
                self.__stdout += out.decode('utf-8')
            else:
                self.__stdout += out
#                     logger.debug("[CMD OUT] | " + out)
        err = self.__process.stderr.read()
        if len(err) > 0:
            if sys.version_info < (3,0):
                self.__stderr += err.decode('utf-8')
            else:
                self.__stderr += err
#                     logger.error("[CMD ERR] | " + out)
                
    def __raise_exception_if_stderr(self):
        if len(self.stderr) > 0:
            msg = "Standard Error:\n{}\n{}\n{}".format('<'*7, self.stderr, '>'*7)
            if len(self.stdout) > 0:
                msg += "\nStandard Output:\n{}\n{}\n{}".format('<'*7, self.stdout, '>'*7)
            raise FunctionalException(f"Error while executing command [{self.cmd}]:\n{msg}")
        
    def kill(self):
        # Note: kill is equivalent to terminate in subprocess implementation
        if self.state == CommandStates.Running:
            self.__process.kill() 
                       
    def terminate(self):
        if self.state == CommandStates.Running:
            self.__process.terminate()
    
    def stop(self):
        if self.state == CommandStates.Running:
            if Tools.do_log(logger, logging.DEBUG):
                logger.debug(f"Stopping command [{self.cmd}] with signal {self.__stop_signal} and PID {self.__process.pid}")
            
            MultitaskManager.kill_process(self.__process.pid, sig=self.__stop_signal, do_kill_children=True, recursively=True)
    
    def raise_error_if_not_success(self, raised_type=FunctionalException):
        if self.state is not CommandStates.Success:
            raise raised_type(f"Error while executing command [{self.cmd}]: [{self.stderr}]")
        
    def __repr__(self):
        return pprint.pformat({'cmd' : self.cmd,
                               'is alive' : self.is_alive() })
            



#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

from builtins import str
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_core.common.tools.tools import Tools
from holado.common.context.session_context import SessionContext
import logging
from holado_value.common.tools.value_types import ValueTypes
from functools import total_ordering
from holado_scripting.common.tools.evaluate_parameters import EvaluateParameters
from holado.holado_config import Config
from holado.common.handlers.undefined import undefined_value, undefined_argument, to_be_defined, not_applicable, is_undefined
from holado_python.standard_library.typing import Typing

logger = logging.getLogger(__name__)


@total_ordering
class Value(object):
    
    def __init__(self, original_value=undefined_argument, value=undefined_argument, do_eval_once=True, eval_params=undefined_argument):
        """
        @summary: Constructor
        @param original_value: Original value (before value expression evaluation) - can be a string or any object
        @param value: Value - can be a string or any object
        """
        self.__original_value = to_be_defined
        self.__value = to_be_defined
        self.__value_type = None
        
        # Manage value evaluation
        self.__value_to_eval = to_be_defined
        self.__evaluated_value = to_be_defined
        self.__evaluated_value_type = to_be_defined
        # By default, raise on interpret and variable evaluation error, but not on eval error
        self.__eval_params_default = eval_params if eval_params is not undefined_argument else EvaluateParameters.default().with_raise_on_eval_error(False)
        self.__do_eval_once = do_eval_once
        
        self.__extract_value_information(original_value, value)
        if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
            logger.trace(f"New Value: original_value=[{self.__original_value}] ; type=[{self.__value_type.name}] ; value=[{self.__value}] ; value_to_eval=[{self.__value_to_eval}]")
        
    def _verify_valid_compared_object(self, other, raise_exception=True):
        res = isinstance(other, Value)
        if not res and raise_exception:
            return TechnicalException(f"Unmanaged comparison between types '{self.__class__.__name__}' and '{other.__class__.__name__}'")
        
    def __eq__(self, other):
        self._verify_valid_compared_object(other)
        if isinstance(other, Value):
            return (self.get_value() == other.get_value())
        else:
            return TechnicalException(f"Unmanaged comparison between types '{self.__class__.__name__}' and '{other.__class__.__name__}'")
    
    def __lt__(self, other):
        self._verify_valid_compared_object(other)
        if isinstance(other, Value):
            return (self.get_value() < other.get_value())
        else:
            return TechnicalException(f"Unmanaged comparison between types '{self.__class__.__name__}' and '{other.__class__.__name__}'")
    
    @property
    def do_eval_once(self):
        return self.__do_eval_once
    
    @property
    def string_value(self):
        value = self.get_value()
        
        if value in [undefined_value, not_applicable]:
            return Config.NOT_APPLICABLE_SYMBOL
        if value is None:
            return Config.NONE_SYMBOL
        else:
            return str(value)
    
    @property
    def original_value(self):
        """
        @summary: Return the original value
        """
        return self.__original_value
    
    @property
    def original_value_type(self):
        """
        @summary: Return the original value type
        """
        return self.__value_type
    
    @property
    def value_type(self):
        return self.get_value_type()
    
    @property
    def value_before_eval(self):
        """
        @summary: Return the value before evaluation
        """
        return self.__value_to_eval
    
    @property
    def value(self):
        """
        @summary: Return the value, after evaluation if needed
        """
        return self.get_value(raise_if_undefined=True)
    
    def get_value(self, raise_if_undefined=False, eval_params=undefined_argument):
        if self.__value_to_eval is not undefined_value:
            self.__evaluate_value_if_needed(eval_params=eval_params)
            res = self.__evaluated_value
        else:
            res = self.__value
        
        if is_undefined(res) and raise_if_undefined:
            msg = "Value is undefined"
            if self.value_type == ValueTypes.NotApplicable:
                msg += " (Not Applicable)"
            raise ValueError(f"Value is undefined (value type: {self.value_type.name} ; original value: [{self.original_value}] (type: {Typing.get_object_class_fullname(self.original_value)}))")
        return res
    
    def get_value_type(self, eval_params=undefined_argument):
        if self.__value_to_eval is not undefined_value:
            self.__evaluate_value_if_needed(eval_params=eval_params)
            res = self.__evaluated_value_type
        else:
            res = self.__value_type
        return res
    
    def __evaluate_value_if_needed(self, eval_params=undefined_argument):
        if self.__evaluated_value is to_be_defined or not self.__do_eval_once:
            eval_params = eval_params if eval_params is not undefined_argument else self.__eval_params_default
            if ValueTypes.is_string(self.__value_type):
                eval_params = eval_params.with_result_type(str)
            self.__evaluated_value_type, self.__evaluated_value = self.__get_expression_evaluator().evaluate_expression_of_information(self.__value_type, self.__value_to_eval, eval_params=eval_params, log_level=logging.TRACE)
        
    def represent(self, indent = 0, do_evaluation = False):
        res_list = []
        
        res_list.append(Tools.get_indent_string(indent))
        
        if ValueTypes.is_string(self.__value_type):
            if do_evaluation:
                res_list.append("'")
                res_list.append(self.string_value)
                res_list.append("'")
            else:
                res_list.append(str(self.original_value))
        elif self.__value_type in [ValueTypes.Boolean, ValueTypes.Integer, ValueTypes.Float, ValueTypes.Null, ValueTypes.NotApplicable, ValueTypes.Symbol, ValueTypes.Generic]:
            if do_evaluation:
                res_list.append(self.string_value)
            else:
                res_list.append(str(self.original_value))
        else:
            raise TechnicalException("Unmanaged value type '{}'".format(self.__value_type.name))
        
        return "".join(res_list)
        
    def __extract_value_information(self, original_value, value):
        # Define original value
        if original_value is not undefined_argument and not (original_value in [undefined_value, None] and value not in [undefined_argument, undefined_value, None]):
            self.__original_value = original_value
        elif value is not undefined_argument:
            if value is not None and isinstance(value, str):
                self.__original_value = f"'{value}'"
            else:
                self.__original_value = value
        
        # Define value information from original value
        self.__value_type, self.__value, self.__value_to_eval = self.__get_expression_evaluator().extract_expression_information(self.original_value)
        if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
            logger.trace(f"extracted value info: {self.__value_type=}, {self.__value=}, {self.__value_to_eval=}")
        
    def __get_expression_evaluator(self):
        from holado_test.behave.context.behave_session_context import BehaveSessionContext

        if BehaveSessionContext.has_instance_scenario_context():
            return SessionContext.instance().get_scenario_context().get_expression_evaluator()
        else:
            return SessionContext.instance().expression_evaluator
        




#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from builtins import object
from holado_core.common.tables.comparators.string_table_cell_comparator import StringTableCellComparator
import re
from holado_core.common.exceptions.functional_exception import FunctionalException
from holado_core.common.tables.comparators.integer_table_cell_comparator import IntegerTableCellComparator
from holado_core.common.tables.comparators.float_table_cell_comparator import FloatTableCellComparator
from holado_core.common.tables.comparators.bytes_table_cell_comparator import BytesTableCellComparator
from holado_core.common.tables.comparators.datetime_table_cell_comparator import DatetimeTableCellComparator
from holado_core.common.tables.comparators.boolean_table_cell_comparator import BooleanTableCellComparator
from holado_core.common.tools.converters.converter import Converter
from holado_core.common.tools.tools import Tools

logger = logging.getLogger(__name__)


class TableComparatorManager(object):
    
    @classmethod
    def convert_compare_method_2_TableCellComparator(cls, compare_method):
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug("Converting compare method name to associated TableCellComparator")

        if compare_method == "Boolean":
            return BooleanTableCellComparator()
        elif compare_method == "Bytes":
            return BytesTableCellComparator()
        elif compare_method == "Datetime":
            return DatetimeTableCellComparator()
        elif compare_method.startswith("Float"):
            regex = re.compile(r"Float\((.*)\)")
            r_match = regex.match(compare_method)
            if r_match:
                precision_str = r_match.group(1)
                if Converter.is_float(precision_str):
                    precision = float(precision_str)
                    return FloatTableCellComparator(diff_precision=precision, relative_precision=None)
                else:
                    raise FunctionalException("The parameter of compare method 'Float' is a precision that must be a float")
            else:
                return FloatTableCellComparator()
        elif compare_method == "Integer":
            return IntegerTableCellComparator()
        elif compare_method == "String":
            return StringTableCellComparator()
        
        



#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

from holado_core.common.finders.tools.enums import FindType



class FindUpdater(object):
    """ Find updater
    """
    
    def __init__(self):
        self.__find_type = FindType.Undefined
        self.__context_values = {}
        self.__parameters_values = {}

    @property
    def find_type(self):
        """
        @return Specific find method type. 
        """
        return self.__find_type
    
    @find_type.setter
    def find_type(self, find_type):
        """
        Set a specific find method type.
        If previously set to Custom, the set is skipped, to preserve custom find type of Finder.
        @param find_type Find method type.
        """
        if self.find_type != FindType.Custom:
            self.__find_type = find_type

    def has_context_value(self, name):
        """
        @param name Property name 
        @return If has context property value
        """
        return name in self.__context_values
    
    def get_context_value(self, name):
        """
        @param name Property name 
        @return Context property value
        """
        return self.__context_values[name]
    
    def set_context_value(self, name, value):
        """
        @param name Property name 
        @param value Property value
        """
        self.__context_values[name] = value
    
    def set_update_root(self, update_root):
        """
        Set if root container must be updated with current container.
        @param update_root boolean
        """
        self.set_context_value("update_root_container", update_root)
    
    def has_parameters_value(self, name):
        """
        @param name Property name 
        @return If has parameters property value
        """
        return name in self.__parameters_values
    
    def get_parameters_value(self, name):
        """
        @param name Property name 
        @return Parameters property value
        """
        return self.__parameters_values[name]
    
    def set_parameters_value(self, name, value):
        """
        @param name Property name 
        @param value Property value
        """
        self.__parameters_values[name] = value
    
    
    
    
    


#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of self software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and self permission notice shall be included in all copies or substantial portions of the Software.
#
# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from holado_core.common.exceptions.technical_exception import TechnicalException
import re
from holado_ui_selenium.ui.gui.selenium.handlers.enums import BrowserTypes
from selenium import webdriver
from selenium.webdriver.common.desired_capabilities import DesiredCapabilities
import tempfile
from holado_system.system.global_system import GlobalSystem, OSTypes
import packaging
from holado_core.common.tools.tools import Tools

logger = logging.getLogger(__name__)



class WebDriverManager(object):
    """ Manager of web drivers.
    """
    chromedriver_download_url = "http://chromedriver.storage.googleapis.com"
    geckodriver_download_url = "https://github.com/mozilla/geckodriver/releases"
    
    def __init__(self, browser_type, path_manager):
        self.__browser_type = browser_type
        self.__browser_version = None
        self.__browser_driver_version = None
        self.__current_proxy = None
    
    @property
    def browser_type(self):
        """
        @return the browser type
        """
        return self.__browser_type
    
    @property
    def current_proxy(self):
        """
        @return Current proxy
        """
        return self.__current_proxy
    
    @current_proxy.setter
    def current_proxy(self, proxy):
        """
        Update current proxy
        @param proxy the proxy to set
        """
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"current proxy has changed ({self.currentProxy} -> {proxy})")
        self.__current_proxy = proxy
    
    @property
    def browser_version(self):
        if self.__browser_version is None:
            self.__browser_version = self.__extract_browser_version()
        
        return self.__browser_version
    
    @property
    def browser_driver_version(self):
        # if self.__browser_driver_version is None:
        #     b_version = self.browser_version
        #     bd_version = self.__get_browser_driver_version_from_db(self.browser_type, b_version)
        #     if bd_version is None:
        #         bd_version = self.__update_browser_driver_dependencies(self.browser_type, b_version)
        #     self.__browser_driver_version = bd_version
        
        return self.__browser_driver_version
    
    def __extract_browser_version(self):
        if GlobalSystem.get_os_type() == OSTypes.Linux:
            if self.browser_type == BrowserTypes.Chrome:
                cmd_res = GlobalSystem.execute_command("chromium --version", do_log_output=False, do_raise_on_stderr=True)
            elif self.browser_type == BrowserTypes.Firefox:
                cmd_res = GlobalSystem.execute_command("firefox --version", do_log_output=False, do_raise_on_stderr=True)
            else:
                raise TechnicalException(f"Unmanaged browser type '{self.browser_type}'")
            
            m = re.search(r"\d+\.[^ ]*", cmd_res.output)
            if m:
                return packaging.version.parse(m.group())
            else:
                raise TechnicalException(f"Failed to find Firefox version in [{cmd_res.output}]")
        else:
            raise TechnicalException(f"Unmanaged OS type '{GlobalSystem.get_os_type().name}'")
        
    # def __update_browser_driver_dependencies(self, browser_type, browser_version):
    #     res = None
    #
    #     # Extract all browser driver versions
    #     bd_versions = self.__extract_browser_driver_versions()
    #
    #     # Find browser driver version adapted to given browser version
    #     for ver in bd_versions:
    #         if browser_version == ver:
    #             res = ver
    #             break
    #         elif browser_version < ver:
    #             continue
    #         else:
    #             if res is None or res < ver:
    #                 res = ver
    #
    #     # Download driver
    #     File tempDir = PathManager.createTemporaryDirectory("webdriver")
    #     String filePath = downloadBrowserDriver(res, tempDir)
    #     String destPath = getPathManager().getChromedriverPath(res, false)
    #     PathManager.makeDirectoriesOfFile(destPath)
    #     PathManager.extractArchiveZip(new File(filePath), new File(destPath).getParentFile())
    #
    #     # Store version in DB
    #     setBrowserDriverVersionInDB(browser_type, browser_driver_version, res)
    #
    #     return res
    #
    #
    # def __downloadBrowserDriver(self, Version browserdriverVersion, File destDir):
    #     switch (browser_type):
    #     case Chrome:
    #         return downloadChromedriver(browserdriverVersion, destDir)
    #     default:
    #         raise TechnicalException(f"Unmanaged browser type '%s'", browser_type))
    #
    #
    #
    # def __downloadChromedriver(self, Version browserdriverVersion, File destDir):
    #     String hostname = getTestProperties().getProperty("browser.chromedriver.download.hostname", None)
    #     String archiveFileName = f"chromedriver_%s.zip", getPathManager().getChromedriverOSName())
    #     String url = f"http:#%s/%s/%s", hostname, browserdriverVersion.getVersion(), archiveFileName)
    #
    #     String res = PathManager.buildPath(destDir.toString(), archiveFileName)
    #     PathManager.makeDirectoriesOfFile(res)
    #     String parameters = f"-f -o \"%s\" %s", res, url)   
    #     CurlCommandResult cmdRes
    #     try:
    #         cmdRes = CommonSystem.executeCurl(parameters, None, None)
    #     } catch (FunctionalException e):
    #         raise TechnicalException(e)
    #
    #     if (cmdRes.hasError()):
    #         String errorMsg = f"Error when downloading file '%s': %s", url, cmdRes.getError())
    #         raise FunctionalException(errorMsg)
    #
    #
    #     return res
    #
    #
    # def __extract_browser_driver_versions(self, ):
    #     switch (browser_type):
    #     case Chrome:
    #         return extractChromedriverVersions()
    #     default:
    #         raise TechnicalException(f"Unmanaged browser type '%s'", browser_type))
    #
    #
    #
    # def __extractChromedriverVersions(self, ):
    #     List<String> res = new ArrayList<String>()
    #
    #     String hostname = getTestProperties().getProperty("browser.chromedriver.download.hostname", None)
    #     CommandResult cmdRes
    #     try:
    #         cmd_res = CommonSystem.execute_curl(f"--location {self.chromedriver_download_url}")
    #     } catch (FunctionalException e):
    #         raise TechnicalException(e)
    #
    #     if cmd_res.has_error:
    #         raise FunctionalException(f"Error when extracting all chrome driver versions with curl: {cmd_res.error}")
    #
    #     String osName = getPathManager().getChromedriverOSName()
    #     Pattern pattern = Pattern.compile(f"<Key>(\\d{2,}[^</]+)/chromedriver_%s.zip</Key>", osName))
    #     Matcher match = pattern.matcher(cmdRes.getOutput())
    #     while (match.find())
    #         res.add(match.group(1))
    #
    #     return res
    #
    #
    # def __get_browser_driver_version_from_db(self, BrowserTypes browser_type, Version browserVersion):
    #     String dbPath = getPathManager().getWebDriverDBPath()
    #     WebDriverDBDriver dbDriver = new WebDriverDBDriver(dbPath)
    #     dbDriver.connect()
    #     try:
    #         String result = dbDriver.getBrowserDriverVersion(browser_type.getValue(), browserVersion.getVersion())
    #         if result is None:
    #             return None
    #         else
    #             return new Version(result)
    #     } finally:
    #         dbDriver.disconnect()
    #
    #
    #
    # def __setBrowserDriverVersionInDB(self, BrowserTypes browser_type, Version browserVersion, Version browserDriverVersion):
    #     String dbPath = getPathManager().getWebDriverDBPath()
    #     WebDriverDBDriver dbDriver = new WebDriverDBDriver(dbPath)
    #     dbDriver.connect()
    #     try:
    #         dbDriver.setBrowserDriverVersion(browser_type.getValue(), browserVersion.getVersion(), browserDriverVersion.getVersion())
    #     } finally:
    #         dbDriver.disconnect()
        
    

    # """
    # @return New WebDriver instance 
    # """
    # def createWebDriver_WithRedo(self, ):
    #     Redo<WebDriver> redo = new Redo<WebDriver>("create web driver"):
    #         @Override
    #         def _process(self, ):
    #             return createWebDriver()
    #
    #
    #         @Override
    #         def _processAfterInterrupt(self, Thread thread):
    #             logger.warn("Creation of web driver has timed out  killing existing browser drivers and retry")
    #             # Kill existing browser driver
    #             try:
    #                 killBrowserDriver()
    #             } catch (FunctionalException exc):
    #                 Tools.logError(logger, "Failed to kill browser driver", exc, false)
    #
    #
    #
    #     }
    #     redo.withProcessTimeout(120, ChronoUnit.SECONDS)
    #     redo.withTimeout(600, ChronoUnit.SECONDS)
    #     return redo.execute()
    
    
    def create_web_driver(self):
        """
        @return New WebDriver instance 
        """
        return self.__create_web_driver(self.browser_type)
    
    def __create_web_driver(self, browser_type):
        if browser_type == BrowserTypes.Chrome:
            return self.__create_web_driver_chrome()
        elif browser_type == BrowserTypes.Firefox:
            return self.__create_web_driver_firefox()
        else:
            raise TechnicalException(f"Unmanaged browser type '{browser_type}'")
                
    def __create_web_driver_chrome(self):
        # For all possible parameters (official site): http://chromedriver.chromium.org/capabilities
        # For possible arguments of ChromeOptions.add_argument: https://peter.sh/experiments/chromium-command-line-switches/
        # For possible options: https://chromium.googlesource.com/chromium/src/+/master/chrome/common/chrome_switches.cc
        # For possible prefs: https://chromium.googlesource.com/chromium/src/+/master/chrome/common/pref_names.cc
        
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug("Creating chrome driver...")

        # Set chromedriver path
        driver_path = self._selenium_path_manager.get_chromedriver_path()
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"Using chromedrive path '{driver_path}'")
        service = webdriver.ChromeService(executable_path=driver_path)
        
        options = webdriver.ChromeOptions()
        preferences = {}
        
        options.unhandled_prompt_behavior = "ignore"
        
        # Set default page to blank page
        preferences["homepage_is_newtabpage"] = True
        preferences["homepage"] = "about:blank"
        
        # Set user profile
        profile_path = tempfile.mkdtemp(prefix="chrome_profile")
        options.add_argument("--user-data-dir=" + profile_path)
        
        # Disable info bar
        options.add_argument("--disable-infobars")
        
        # Adjust popup settings
        options.add_argument("--disable-popup-blocking")
        
        # Prompt for download
        preferences["download.prompt_for_download"] = True
        
        # Enable Flash
        preferences["profile.default_content_setting_values.plugins"] = 1
        preferences["profile.content_settings.plugin_whitelist.adobe-flash-player"] = 1
        preferences["profile.content_settings.exceptions.plugins.*,*.per_resource.adobe-flash-player"] = 1
        
        options.add_experimental_option("prefs", preferences)
        
        res = webdriver.Chrome(options=options, service=service)
        
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug("Created chrome driver.")
        
        return res
        
    def __create_web_driver_firefox(self):
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug("Creating firefox driver...")

        # Set geckodriver path
        driver_path = self._selenium_path_manager.get_geckodriver_path()
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"Using geckodriver path '{driver_path}'")
        service = webdriver.FirefoxService(executable_path=driver_path)
        
        firefox_capabilities = DesiredCapabilities.FIREFOX
        firefox_capabilities['marionette'] = True

        options = webdriver.FirefoxOptions()
        options.set_preference("webdriver.log.file", "/dev/None")

        res = webdriver.Firefox(options=options, service=service)
        
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug("Created firefox driver.")
        
        return res
    
    def kill_browser_driver(self):
        """
        Kill browser driver
        """
        self.__kill_browser_driver(self.browser_type)
    
    
    def __kill_browser_driver(self, browser_type):
#         switch (CommonSystem.getOSType()):
#         case Windows:
#             Map<Integer, String> processes = CommonSystem.getRunningProcesses()
#
#             switch (browser_type):
#             case Chrome:
#                 CommonSystem.closeProcessesForCommandPattern(processes, "chromedriver.exe", CloseParameters.kill())
#                 CommonSystem.closeProcessesForCommandPattern(processes, "chrome.exe", CloseParameters.kill())
#                 break
#
#             case Firefox:
#                 CommonSystem.closeProcessesForCommandPattern(processes, "geckodriver.exe", CloseParameters.kill())
#                 CommonSystem.closeProcessesForCommandPattern(processes, "firefox.exe", CloseParameters.kill())
#                 break
#
#             case IE:
#                 CommonSystem.closeProcessesForCommandPattern(processes, "IEDriverServer.exe", CloseParameters.kill())
#                 CommonSystem.closeProcessesForCommandPattern(processes, "iexplore.exe", CloseParameters.kill())
#                 break
#
#             default:
#                 raise TechnicalException(f"Unmanaged browser type '%s'", browser_type))
#
#             break
#
# #        case Linux:
# #            break
#
#         default:
#             raise NotImplementedException(f"Method not implemented for OS type '%s'", CommonSystem.getOSType().name()))
#
#
#         # Wait to let time for OS to release resources
#         Tools.sleep(10000)
    
    
    
    
    
    



#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
import copy
import weakref
import behave
from holado.common.context.session_context import SessionContext
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_multitask.multitasking.multitask_manager import MultitaskManager
import time

logger = logging.getLogger(__name__)


##### Implementation working but not working when launching behave in parallel in multiple threads

# class BehaveManager(object):
#     """
#     Manage Behave features.
#     """
#
#     def __init__(self):
#         self.__main_pid = None
#         self.__main_context_by_pid = {}
#         self.__context_by_pid_and_tid = {}
#
#
#     def clear(self):
#         # TODO EKL: clear only current (pid, tid), since behave can be launched in parallel threads
#         # pid = os.getpid()
#         # if pid in self.__main_context_by_pid:
#         #     del self.__main_context_by_pid[pid]
#         # if pid in self.__context_by_pid_and_tid:
#         #     del self.__context_by_pid_and_tid[pid]
#         pass
#
#     def has_main_pid(self):
#         return self.__main_pid is not None
#
#     def get_main_pid(self):
#         return self.__main_pid
#
#     def get_main_context(self, pid=None, raise_exception=True):
#         context_info = self.__get_main_context_info(pid, raise_exception)
#         if context_info is None:
#             return None
#         else:
#             return context_info[0]
#
#     def __get_main_context_info(self, pid=None, raise_exception=True):
#         if pid is None:
#             pid = os.getpid()
#         if pid not in self.__main_context_by_pid:
#             if raise_exception:
#                 raise TechnicalException("Main behave context is not defined")
#             else:
#                 return None
#         return self.__main_context_by_pid[pid]
#
#     def set_main_context(self, context, runner=None, raise_if_already_exists=True):
#         if self.get_main_context(raise_exception=False) is not None:
#             if self.__is_main_context_with_indendant_runner(raise_exception=False):
#                 SessionContext.instance().after_scenario(SessionContext.instance().get_scenario_context().scenario)
#                 SessionContext.instance().after_feature(SessionContext.instance().get_feature_context().feature)
#             else:
#                 if raise_if_already_exists:
#                     raise TechnicalException(f"Main behave context is already set: [{self.get_main_context(raise_exception=False)}]")
#                 else:
#                     logger.warning(f"Resetting main behave context: [{self.get_main_context(raise_exception=False)}] -> [{context}]")
#
#         pid = os.getpid()
#         thread_id = MultitaskManager.get_thread_id()
#
#         if self.__main_pid is None:
#             self.__main_pid = pid
#             # logger.print(f"+++++ set main pid: {self.__main_pid}")
#
#         self.__main_context_by_pid[pid] = (context, runner)
#         if pid not in self.__context_by_pid_and_tid:
#             self.__context_by_pid_and_tid[pid] = {}
#         self.__context_by_pid_and_tid[pid][thread_id] = (context, runner)
#         # logger.print(f"+++++ set main context: pid={self.__main_pid} ; tid={thread_id} ; {context=} ; {runner=}")
#
#     def __is_main_context_with_indendant_runner(self, raise_exception=True):
#         import holado_test.behave.independant_runner
#
#         context_info = self.__get_main_context_info(raise_exception=raise_exception)
#         if context_info is None:
#             return False
#
#         runner = context_info[1] if context_info[1] is not None else context_info[0]._runner
#         if not isinstance(runner, behave.runner.Runner):
#             raise TechnicalException(f"Wrong runner instance is verified")
#
#         return isinstance(runner, holado_test.behave.independant_runner.IndependantRunner)
#
#     def use_independant_runner(self, step_paths=None):
#         """
#         Use this method to run an independant behave runner,
#         so that a main context is set that can be used to execute steps
#         with method "execute_steps" without executing behave features.
#
#         WARNING: Current implementation has limitations.
#                  Current working directory must contain:
#                     - a folder "features" with at least one .feature file, even empty
#                     - a folder "steps" with a x_steps.py file importing all step files
#                     - a file "environment.py" containing "from XXX.behave_environment import *"
#         """
#         ### Current working implementation
#         from holado_test.behave.behave_function import BehaveFunction
#         from holado_multitask.multithreading.functionthreaded import FunctionThreaded
#         import time
#         behave_args = f'--no-source --no-skipped --format null --no-summary --no-capture --no-capture-stderr --no-logcapture -r holado_test.behave.independant_runner:IndependantRunner'
#         behave_func = BehaveFunction(behave_args)
#         thread = FunctionThreaded(behave_func.run, args=[], name=f"Thread with fake behave run")
#         thread.start()
#         time.sleep(1)
#
#         ### Other tried implementation:
#
#         # from holado_multitask.multithreading.functionthreaded import FunctionThreaded
#         # import time
#         #
#         # behave_args = f'--no-source --no-skipped --format null --no-summary --no-capture --no-capture-stderr --no-logcapture -r holado_test.behave.independant_runner:IndependantRunner'
#         # def run(behave_args):
#         #     from behave.configuration import Configuration
#         #     from behave.runner_plugin import RunnerPlugin
#         #     from behave.runner_util import reset_runtime
#         #
#         #     # Copy of useful code in behave.__main__.main
#         #     config = Configuration(behave_args)
#         #     if not config.format:
#         #         config.format = [config.default_format]
#         #     elif config.format and "format" in config.defaults:
#         #         # -- CASE: Formatter are specified in behave configuration file.
#         #         #    Check if formatter are provided on command-line, too.
#         #         if len(config.format) == len(config.defaults["format"]):
#         #             # -- NO FORMATTER on command-line: Add default formatter.
#         #             config.format.append(config.default_format)
#         #
#         #     reset_runtime()
#         #     runner = RunnerPlugin().make_runner(config, step_paths=step_paths)
#         #     runner.run()
#         # thread = FunctionThreaded(run, args=[behave_args], name=f"Thread with independant behave run")
#         # thread.start()
#         # time.sleep(1)
#
#         ### Other tried implementation:
#
#         # context = behave.runner.Context(runner)
#         # runner.context = context
#         #
#         # runner.load_hooks()
#         # runner.load_step_definitions()
#         # feature_locations = [filename for filename in runner.feature_locations()
#         #                      if not runner.config.exclude(filename)]
#         # features = runner.parse_features(feature_locations, language=self.config.lang)
#         # runner.features.extend(features)
#         # runner.feature = runner.features[0]
#         #
#         # self.set_main_context(context, runner=runner)
#
#
#     def get_context(self, raise_exception=True):
#         context_info = self.__get_context_info(raise_exception)
#         if context_info is None:
#             return None
#         else:
#             return context_info[0]
#
#     def __get_context_info(self, raise_exception=True):
#         pid = os.getpid()
#         if pid not in self.__context_by_pid_and_tid:
#             # Initialize main context of this pid as a sub context of main pid
#             # raise TechnicalException(f"Process {pid} doesn't exist in __context_by_pid_and_tid")
#             new_context, new_runner = self.__new_context_runner(pid_ref=self.__main_pid, raise_exception=raise_exception)
#             self.set_main_context(new_context, new_runner)
#
#         tid = MultitaskManager.get_thread_id()
#         if tid not in self.__context_by_pid_and_tid[pid]:
#             new_context, new_runner = self.__new_context_runner(raise_exception)
#             # new_context.thread_id = tid
#             # logger.debug(f"New context ({id(new_context)}) for thread {tid}")
#
#             self.__context_by_pid_and_tid[pid][tid] = (new_context, new_runner)
#
#         res = self.__context_by_pid_and_tid[pid][tid]
#         # self.__verify_context(res)
#         return res
#
#     def __new_context_runner(self, raise_exception, pid_ref=None):
#         ref_context = self.get_main_context(pid=pid_ref, raise_exception=raise_exception)
#
#         # Create a copy of the runner
#         res_run = copy.copy(ref_context._runner.__repr__.__self__)   # Context._runner is a weakref.proxy
#
#         # Create new context
#         res_con = behave.runner.Context(res_run)
#         res_con.feature = copy.copy(ref_context.feature)
#         res_con.feature.parser = copy.copy(ref_context.feature.parser)
#
#         # Update cross references
#         res_run.context = res_con
#         res_con._runner = weakref.proxy(res_run)    # Context._runner is a weakref.proxy
#
#         # Update context with attributes used by HolAdo
#         self.before_scenario(res_con)
#
#         return res_con, res_run
#
#     def before_scenario(self, context=None):
#         if context is None:
#             context = self.get_main_context()
#
#         # Initialize step information
#         context.sub_step_exception = None
#
#         # Manage expected exception
#         context.expected_exception_str = None
#         context.expected_exception_pattern = None
#         context.in_expected_exception_step = False
#         context.expected_exception_step_level = -1
#         context.expected_exception_thread_id = None
#
#     # def __log_registered_contexts(self, name=None):
#     #     msg_list = []
#     #     if name:
#     #         msg_list.append(f"[{name}] Registered contexts:")
#     #     else:
#     #         msg_list.append("Registered contexts:")
#     #     msg_list.append(f"    Main context: id={id(self.__main_context)} (thread: {self.__main_context.thread_id})")
#     #     for tid, con_run in self.__context_by_pid_and_tid.items():
#     #         context, _ = con_run
#     #         msg_list.append(f"    Context of thread {tid}: id={id(context)} (thread: {context.thread_id})")
#     #     logger.debug("\n".join(msg_list))
#
#     # def __verify_context(self, context):
#     #     thread_id = MultitaskManager.get_thread_id()
#     #     if context.thread_id != thread_id:
#     #         raise TechnicalException(f"Current thread and context thread are different (current thread: {thread_id} ; context thread: {context.thread_id} ; context: {id(context)})")
#






##### Implementation working when launching behave in parallel in multiple threads

class BehaveManager(object):
    """
    Manage Behave features.
    """
    
    def __init__(self):
        self.__context_by_pid_and_tid = {}
    
    def clear(self):
        pid = MultitaskManager.get_process_id()
        tid = MultitaskManager.get_thread_id()
        if pid in self.__context_by_pid_and_tid:
            if tid in self.__context_by_pid_and_tid[pid]:
                del self.__context_by_pid_and_tid[pid][tid]
                # TODO: delete related threads
            if len(self.__context_by_pid_and_tid[pid]) == 0:
                del self.__context_by_pid_and_tid[pid]
            # TODO: delete related processes
        
    def has_main_pid(self):
        return self.get_main_pid() is not None
    
    def get_main_pid(self):
        if len(self.__context_by_pid_and_tid) == 0:
            return None
        return next(iter(self.__context_by_pid_and_tid))
    
    def get_main_context(self, pid=None, raise_exception=True):
        context_info = self.__get_main_context_info(pid, raise_exception)
        return self.__get_context_from_info(context_info)
    
    def __get_context_from_info(self, context_info):
        return context_info[0] if context_info is not None else None
    
    def __get_runner_from_info(self, context_info):
        return context_info[1] if context_info is not None else None
    
    def __is_reference_from_info(self, context_info):
        return context_info[2] if context_info is not None else None
    
    def __get_main_context_info(self, pid=None, raise_exception=True):
        custom_pid = pid is not None and pid != MultitaskManager.get_process_id()
        if pid is None:
            pid = MultitaskManager.get_process_id()
        
        if pid not in self.__context_by_pid_and_tid or len(self.__context_by_pid_and_tid[pid]) == 0:
            if custom_pid:
                raise TechnicalException(f"Not managed to get main behave context on custom PID {pid}")
            else:
                ppid = MultitaskManager.get_parent_process_id()
            if ppid not in self.__context_by_pid_and_tid or len(self.__context_by_pid_and_tid[ppid]) == 0:
                if raise_exception:
                    raise TechnicalException(f"Main behave context is not defined either for current process PID {pid} and parent process PID {ppid} (existing context: {self.__context_by_pid_and_tid})")
                else:
                    return None
            else:
                # Use same context than parent process
                context, runner, _ = next(iter(self.__context_by_pid_and_tid[ppid].values()))
                tid = MultitaskManager.get_thread_id()
                self.__set_context(pid, tid, context, runner, is_reference=False)
        
        return next(iter(self.__context_by_pid_and_tid[pid].values()))
    
    def set_main_context(self, context, runner=None, raise_if_already_exists=True):
        context_info = self.__get_main_context_info(raise_exception=False)
        if context_info is not None and self.__get_context_from_info(context_info) is not None:
            if not self.__is_reference_from_info(context_info):
                self.__remove_context()
            elif self.__is_context_info_with_indendant_runner(context_info, raise_exception=False):
                if SessionContext.instance().has_scenario_context(is_reference=True):
                    SessionContext.instance().after_scenario(SessionContext.instance().get_scenario_context().scenario)
                if SessionContext.instance().has_feature_context(is_reference=True):
                    SessionContext.instance().after_feature(SessionContext.instance().get_feature_context().feature)
            else:
                if raise_if_already_exists:
                    raise TechnicalException(f"Main behave context is already set: [{self.get_main_context(raise_exception=False)}]")
                else:
                    logger.warning(f"Resetting main behave context: [{self.get_main_context(raise_exception=False)}] -> [{context}]")
        
        # Initialize context with HolAdo needs
        self.__init_context(context)
        
        pid = MultitaskManager.get_process_id()
        tid = MultitaskManager.get_thread_id()
        self.__set_context(pid, tid, context, runner, is_reference=True)
        # logger.print(f"+++++ set main context: pid={pid} ; tid={tid} ; {context=} ; {runner=}")
    
    def __is_context_info_with_indendant_runner(self, context_info, raise_exception=True):
        import holado_test.behave.independant_runner
        
        runner = self.__get_runner_from_info(context_info)
        if runner is None:
            runner = self.__get_context_from_info(context_info)._runner
            
        if not isinstance(runner, behave.runner.Runner):
            raise TechnicalException(f"Wrong runner instance is verified")
        
        return isinstance(runner, holado_test.behave.independant_runner.IndependantRunner)
    
    def use_independant_runner(self, step_paths=None):
        """
        Use this method to run an independant behave runner,
        so that a main context is set that can be used to execute steps
        with method "execute_steps" without executing behave features.
        
        WARNING: Current implementation has limitations.
                 Current working directory must contain:
                    - a folder "features" with at least one .feature file, even empty
                    - a folder "steps" with a x_steps.py file importing all step files
                    - a file "environment.py" containing "from XXX.behave_environment import *"
        """
        ### Current working implementation
        from holado_test.behave.behave_function import BehaveFunction
        from holado_multitask.multithreading.functionthreaded import FunctionThreaded
        from holado_core.common.handlers.wait import WaitFuncResult
        
        # Run behave with runner IndependantRunner in a thread
        behave_args = f'--no-source --no-skipped --format null --no-summary --no-capture --no-capture-stderr --no-logcapture -r holado_test.behave.independant_runner:IndependantRunner'
        behave_func = BehaveFunction(behave_args)
        thread = FunctionThreaded(behave_func.run, args=[], name=f"Thread with fake behave run")
        thread.start()
        thread_uid = MultitaskManager.get_thread_uid(thread)
        
        # Use the same thread context for this thread
        SessionContext.instance().multitask_manager.relate_thread_to(MultitaskManager.get_thread_uid(), thread_uid)
        
        # Wait until associated behave context exists
        def context_exists(pid, tid):
            return pid in self.__context_by_pid_and_tid and tid in self.__context_by_pid_and_tid[pid]
        wait_context = WaitFuncResult(f"wait independent behave context exists", 
                                      lambda: context_exists(MultitaskManager.get_process_id(), thread_uid))
        wait_context.redo_until(True)
        wait_context.execute()
        # Note: previous waiting mechanism is sometimes to rapid to execute steps just after
        time.sleep(0.01)
        
        ### Other tried implementation:
        
        # from holado_multitask.multithreading.functionthreaded import FunctionThreaded
        # import time
        #
        # behave_args = f'--no-source --no-skipped --format null --no-summary --no-capture --no-capture-stderr --no-logcapture -r holado_test.behave.independant_runner:IndependantRunner'
        # def run(behave_args):
        #     from behave.configuration import Configuration
        #     from behave.runner_plugin import RunnerPlugin
        #     from behave.runner_util import reset_runtime
        #
        #     # Copy of useful code in behave.__main__.main
        #     config = Configuration(behave_args)
        #     if not config.format:
        #         config.format = [config.default_format]
        #     elif config.format and "format" in config.defaults:
        #         # -- CASE: Formatter are specified in behave configuration file.
        #         #    Check if formatter are provided on command-line, too.
        #         if len(config.format) == len(config.defaults["format"]):
        #             # -- NO FORMATTER on command-line: Add default formatter.
        #             config.format.append(config.default_format)
        #
        #     reset_runtime()
        #     runner = RunnerPlugin().make_runner(config, step_paths=step_paths)
        #     runner.run()
        # thread = FunctionThreaded(run, args=[behave_args], name=f"Thread with independant behave run")
        # thread.start()
        # time.sleep(1)
        
        ### Other tried implementation:
        
        # context = behave.runner.Context(runner)
        # runner.context = context
        #
        # runner.load_hooks()
        # runner.load_step_definitions()
        # feature_locations = [filename for filename in runner.feature_locations()
        #                      if not runner.config.exclude(filename)]
        # features = runner.parse_features(feature_locations, language=self.config.lang)
        # runner.features.extend(features)
        # runner.feature = runner.features[0]
        #
        # self.set_main_context(context, runner=runner)
        
        
    def get_context(self, raise_exception=True):
        context_info = self.__get_context_info(raise_exception)
        return self.__get_context_from_info(context_info)
        
    def __get_context_info(self, raise_exception=True):
        pid = MultitaskManager.get_process_id()
        if pid not in self.__context_by_pid_and_tid:
            # Initialize main context of this pid as a sub context of main pid
            # raise TechnicalException(f"Process {pid} doesn't exist in __context_by_pid_and_tid")
            # new_context, new_runner = self.__new_context_runner(pid_ref=self.__main_pid, raise_exception=raise_exception)
            # self.set_main_context(new_context, new_runner)
            self.__context_by_pid_and_tid[pid] = {}
        
        tid = MultitaskManager.get_thread_id()
        if tid not in self.__context_by_pid_and_tid[pid]:
            # Create a new context using as reference the first context of current process
            new_context, new_runner = self.__new_context_runner(raise_exception, pid_ref=pid)
            # new_context.thread_id = tid
            # logger.debug(f"New context ({id(new_context)}) for thread {tid}")
            self.__set_context(pid, tid, new_context, new_runner, is_reference=True)
            
        res = self.__context_by_pid_and_tid[pid][tid]
        # self.__verify_context(res)
        return res
    
    def __set_context(self, pid, tid, context, runner, is_reference=True):
        if pid not in self.__context_by_pid_and_tid:
            self.__context_by_pid_and_tid[pid] = {}
        self.__context_by_pid_and_tid[pid][tid] = (context, runner, is_reference)
        
    def __remove_context(self, pid=None, tid=None):
        if pid is None:
            pid = MultitaskManager.get_process_id()
        if tid is None:
            tid = MultitaskManager.get_thread_id()
            
        if pid in self.__context_by_pid_and_tid and tid in self.__context_by_pid_and_tid[pid]:
            del self.__context_by_pid_and_tid[pid][tid]
        else:
            raise TechnicalException(f"No context is set for thread ID {tid} and process ID {pid}")
        
    def __new_context_runner(self, raise_exception, pid_ref=None):
        ref_context = self.get_main_context(pid=pid_ref, raise_exception=raise_exception)
        
        # Create a copy of the runner
        res_run = copy.copy(ref_context._runner.__repr__.__self__)   # Context._runner is a weakref.proxy
        
        # Create new context
        res_con = behave.runner.Context(res_run)
        res_con.feature = copy.copy(ref_context.feature)
        res_con.feature.parser = copy.copy(ref_context.feature.parser)
        
        # Update cross references
        res_run.context = res_con
        res_con._runner = weakref.proxy(res_run)    # Context._runner is a weakref.proxy
        
        # Update context with attributes used by HolAdo
        self.before_scenario(res_con)
        
        # Initialize context with HolAdo needs
        self.__init_context(res_con)
        
        return res_con, res_run
    
    def before_scenario(self, context=None):
        if context is None:
            context = self.get_main_context()
            
        self.__reset_context(context)
        
    def __init_context(self, context):
        # Initialize step information
        if not hasattr(context, 'sub_step_exception'):
            context.sub_step_exception = None
        
        # Manage expected exception
        if not hasattr(context, 'expected_exception_str'):
            context.expected_exception_str = None
        if not hasattr(context, 'expected_exception_pattern'):
            context.expected_exception_pattern = None
        if not hasattr(context, 'in_expected_exception_step'):
            context.in_expected_exception_step = False
        if not hasattr(context, 'expected_exception_step_level'):
            context.expected_exception_step_level = -1
        if not hasattr(context, 'expected_exception_thread_id'):
            context.expected_exception_thread_id = None
        
    def __reset_context(self, context):
        # Initialize step information
        context.sub_step_exception = None
        
        # Manage expected exception
        context.expected_exception_str = None
        context.expected_exception_pattern = None
        context.in_expected_exception_step = False
        context.expected_exception_step_level = -1
        context.expected_exception_thread_id = None
        
    # def __log_registered_contexts(self, name=None):
    #     msg_list = []
    #     if name:
    #         msg_list.append(f"[{name}] Registered contexts:")
    #     else:
    #         msg_list.append("Registered contexts:")
    #     msg_list.append(f"    Main context: id={id(self.__main_context)} (thread: {self.__main_context.thread_id})")
    #     for tid, con_run in self.__context_by_pid_and_tid.items():
    #         context, _ = con_run
    #         msg_list.append(f"    Context of thread {tid}: id={id(context)} (thread: {context.thread_id})")
    #     logger.debug("\n".join(msg_list))

    # def __verify_context(self, context):
    #     thread_id = MultitaskManager.get_thread_id()
    #     if context.thread_id != thread_id:
    #         raise TechnicalException(f"Current thread and context thread are different (current thread: {thread_id} ; context thread: {context.thread_id} ; context: {id(context)})")
    
    
    def abort_execution(self, reason):
        self.get_main_context().abort(reason=reason)
    
    def is_execution_aborted(self):
        return self.get_main_context().aborted






#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

from builtins import super
import logging
from holado_test.behave.behave_function import BehaveFunction
from typing import Dict, Union
from holado_core.common.tools.tools import Tools
from holado_core.common.actors.actions import Action
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_python.standard_library.typing import Typing
from holado.common.handlers.undefined import not_applicable

logger = logging.getLogger(__name__)


class ActionInfo(object):
    def __init__(self, uid:str, action:str, params:Dict[str, Union[str,int]], static_params:Dict[str, Union[str,int]], index:int):
        self.uid = uid
        self.action = action
        self.params = params
        self.static_params = static_params
        self.index = index
        
    def __str__(self)->str:
        return f"{Typing.get_object_class_fullname(self)}({self.__dict__})"

class BehaveActionInfo(ActionInfo):
    def __init__(self, uid:str, action:str, params:Dict[str, Union[str,int]], static_params:Dict[str, Union[str,int]], index:int, behave_args:str):
        super().__init__(uid, action, params, static_params, index)
        self.behave_args = behave_args


class FeaturesBehaveAction(Action):
    '''Behave action running features.
    The features to execute are defined by parameter behave_args.
    
    It is possible to pass arguments to the features, by writing the scenarios with expected variables.
    At execution, given parameter 'variable_values' must define expected variable values. 
    '''
    
    def __init__(self, behave_args, variable_manager):
        # Method execute is overriden and doesn't use a method reference, thus parameter func_execute is None
        super().__init__("behave", None)
        self.__behave_function = BehaveFunction(behave_args)
        self.__variable_manager = variable_manager
    
    def execute(self, variable_values=None):
        """Execute features with given variable values.
        """
        if variable_values is not None and not isinstance(variable_values, dict):
            raise TechnicalException(f"Parameter 'variable_values' must be a dict")
        
        if variable_values:
            for key, value in variable_values.items():
                self.__variable_manager.register_variable(key, value)
        
        try:
            return self.__behave_function.run()
        except Exception as exc:
            raise TechnicalException(f"Execution of features with behave failed: {exc}") from exc
    
    
class BehaveActionRunner():
    '''
    Behave action runner.
    Return True if succeeded
    '''
    
    def __init__(self, action_info, variable_manager):
        self.__action = FeaturesBehaveAction(action_info.behave_args, variable_manager)
        self.__action_info = action_info
    
    def __compute_variable_values(self):
        res = {}
        self.__add_variable_values(res, self.__action_info.static_params)
        self.__add_variable_values(res, self.__action_info.params)
        return res
    
    def __add_variable_values(self, res, params):
        if params:
            if isinstance(params, dict):
                for key, value in params.items():
                    self.__add_variable_value(res, key, value)
            elif isinstance(params, list):
                for key, value in params:
                    self.__add_variable_value(res, key, value)
            else:
                raise TechnicalException(f"Unexpected params type '{type(params)}'")
    
    def __add_variable_value(self, res, var_name, value):
        if var_name == 'STATUS':
            return
        if value is not_applicable:
            res[var_name] = 'N/A'
        else:
            res[var_name] = value
    
    def run(self):
        logger.info(f"Processing action '{self.__action_info.action}' of index {self.__action_info.index}: {self.__action_info.params}")
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"Running behave with arguments '{self.__action_info.behave_args}' on action of index {self.__action_info.index}: {self.__action_info.params}")
        
        variable_values = self.__compute_variable_values()
        
        logger.info(f"Launching behave with arguments: {self.__action_info.behave_args}")
        if Tools.do_log(logger, logging.DEBUG):
            if variable_values:
                logger.debug(f"   and with variables:\n{Tools.represent_object(variable_values, 8)}")
        res = self.__action.execute(variable_values)
        logger.info(f"behave finished with result: {res}")
        
        if res:
            logger.print(f"Success of action '{self.__action_info.action}' on action of index {self.__action_info.index}: {self.__action_info.params}")
        else:
            logger.error(f"Action '{self.__action_info.action}' failed on action of index {self.__action_info.index}: {self.__action_info.params}")
                
        return res
    
    



# -*- coding: utf-8 -*-

import holado_test.behave.behave_environment


# Import default hook implementation
from holado_test.behave.behave_environment import *

# Redefine specific hook implementation
def before_all(context):
    holado_test.behave.behave_environment.before_all(context)

def after_all(context):
    holado_test.behave.behave_environment.after_all(context)

def before_feature(context, feature):
    holado_test.behave.behave_environment.before_feature(context, feature)

def after_feature(context, feature):
    holado_test.behave.behave_environment.after_feature(context, feature)

def before_scenario(context, scenario):
    holado_test.behave.behave_environment.before_scenario(context, scenario)

def after_scenario(context, scenario):
    holado_test.behave.behave_environment.after_scenario(context, scenario)

def before_step(context, step):
    holado_test.behave.behave_environment.before_step(context, step)

def after_step(context, step):
    holado_test.behave.behave_environment.after_step(context, step)

    




# -*- coding: utf-8 -*-

#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################


from holado_test.behave.behave import *  # @UnusedWildImport
from holado_test.behave.behave import render_step_table
import copy
from holado_value.common.tools.value_types import ValueTypes
from holado_core.common.tables.table_manager import TableManager
from holado_value.common.tables.comparators.table_2_value_table_comparator import Table2ValueTable_Comparator
from holado_core.common.tools.tools import reversed_enumerate
from holado_test.scenario.step_tools import StepTools
from holado_value.common.tables.comparators.table_2_value_table_with_header_comparator import Table2ValueTable_WithHeaderComparator
from holado_value.common.tables.comparators.table_2_value_table_cell_comparator import Table2ValueTable_CellComparator
from holado_core.common.tables.table_with_header import TableWithHeader
from holado_core.common.tables.table import Table
from holado_core.common.tables.table_row import TableRow
from holado_core.common.tables.comparators.string_table_row_comparator import StringTableRowComparator
import logging
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_core.common.exceptions.functional_exception import FunctionalException
from holado.common.context.session_context import SessionContext
from holado_value.common.tables.value_table import ValueTable
from holado_value.common.tables.value_table_with_header import ValueTableWithHeader
from holado_core.common.tools.converters.converter import Converter
from holado_test.behave.scenario.behave_step_tools import BehaveStepTools
from holado_value.common.tables.converters.value_table_converter import ValueTableConverter
from holado_python.standard_library.typing import Typing
from holado_value.common.tables.value_table_manager import ValueTableManager
from holado_core.common.tables.converters.table_converter import TableConverter

logger = logging.getLogger(__name__)


def __get_scenario_context():
    return SessionContext.instance().get_scenario_context()

def __get_text_interpreter():
    return __get_scenario_context().get_text_interpreter()

def __get_variable_manager():
    return __get_scenario_context().get_variable_manager()

@Given(r"(?P<var_name>{Variable}) = table")
def step_impl(context, var_name):
    var_name = StepTools.evaluate_variable_name(var_name)
    
    table = BehaveStepTools.convert_step_table_2_value_table(context.table)
    
    __get_variable_manager().register_variable(var_name, table)

@Given(r"(?P<var_name>{Variable}) = table with header")
def step_impl(context, var_name):
    var_name = StepTools.evaluate_variable_name(var_name)
    
    table = BehaveStepTools.convert_step_table_2_value_table_with_header(context.table)
    
    __get_variable_manager().register_variable(var_name, table)

@Given(r"(?P<var_name>{Variable}) = object table")
def step_impl(context, var_name):
    var_name = StepTools.evaluate_variable_name(var_name)
    table = BehaveStepTools.convert_step_table_2_value_table(context.table)
    
    res = ValueTableConverter.convert_value_table_2_table(table)
    
    __get_variable_manager().register_variable(var_name, res)

@Given(r"(?P<var_name>{Variable}) = object table with header")
def step_impl(context, var_name):
    var_name = StepTools.evaluate_variable_name(var_name)
    table = BehaveStepTools.convert_step_table_2_value_table_with_header(context.table)
    
    res = ValueTableConverter.convert_value_table_2_table(table)
    
    __get_variable_manager().register_variable(var_name, res)

@Step(r"(?P<var_name>{Variable}) = convert dictionary (?P<obj_str>{Str}) to name/value table")
def step_impl(context, var_name, obj_str):
    var_name = StepTools.evaluate_variable_name(var_name)
    obj = StepTools.evaluate_scenario_parameter(obj_str)
    if isinstance(obj, str):
        obj = eval(obj)
    if not isinstance(obj, dict):
        raise TechnicalException(f"Expecting dict, received {Typing.get_object_class_fullname(obj)}")
    
    table = TableManager.convert_dict_2_name_value_table(obj)
    
    __get_variable_manager().register_variable(var_name, table)

@Step(r"(?P<var_name>{Variable}) = convert dictionary (?P<obj_str>{Str}) to table with keys as columns")
def step_impl(context, var_name, obj_str):
    var_name = StepTools.evaluate_variable_name(var_name)
    obj = StepTools.evaluate_scenario_parameter(obj_str)
    if isinstance(obj, str):
        obj = eval(obj)
    if not isinstance(obj, dict):
        raise TechnicalException(f"Expecting dict, received {Typing.get_object_class_fullname(obj)}")
    
    table = TableManager.convert_dict_2_table_with_keys_as_column(obj)
    
    __get_variable_manager().register_variable(var_name, table)

@Step(r"(?P<var_name>{Variable}) = convert table with header (?P<table>{Variable}) to list of dictionary(?: \((?P<as_generator_str>as generator)\))?")
def step_impl(context, var_name, table, as_generator_str):
    var_name = StepTools.evaluate_variable_name(var_name)
    table = StepTools.evaluate_scenario_parameter(table)
    as_generator = as_generator_str is not None
    if not TableManager.is_table_with_header(table):
        raise TechnicalException(f"Table must be a table with header (obtained type: {Typing.get_object_class_fullname(table)})")
    
    if ValueTableManager.is_value_table(table):
        res = ValueTableConverter.convert_table_with_header_to_dict_list(table, as_generator=as_generator)
    else:
        res = TableConverter.convert_table_with_header_to_dict_list(table, as_generator=as_generator)
    
    __get_variable_manager().register_variable(var_name, res)

@Step(r'(?P<var_name>{Variable}) = convert object (?P<obj_str>{Variable}) to name/value table')
def step_impl(context, var_name, obj_str):
    var_name = StepTools.evaluate_variable_name(var_name)
    obj = StepTools.evaluate_variable_value(obj_str)
    
    table = TableManager.convert_object_attributes_2_name_value_table(obj)
    
    __get_variable_manager().register_variable(var_name, table)
    
#TODO EKL: move step
@Step(r"(?P<var_name>{Variable}) = convert string (?P<list_obj_str>{Variable}) to list")
def step_impl(context, var_name, list_obj_str):
    var_name = StepTools.evaluate_variable_name(var_name)
    list_str = StepTools.evaluate_variable_value(list_obj_str)
    
    list_obj = eval(list_str)
    if not isinstance(list_obj, list):
        raise FunctionalException("Given string is not a string representation of a list")
    
    __get_variable_manager().register_variable(var_name, list_obj)

@Step(r'(?P<var_name>{Variable}) = convert list (?P<list_obj_str>{Variable}) to table with object attributes as columns')
def step_impl(context, var_name, list_obj_str):
    var_name = StepTools.evaluate_variable_name(var_name)
    list_obj = StepTools.evaluate_variable_value(list_obj_str)
    
    table = TableManager.convert_object_list_2_table_with_attributes_as_column(list_obj)
    
    __get_variable_manager().register_variable(var_name, table)

@Step(r'(?P<var_name>{Variable}) = convert list (?P<list_obj_str>{Variable}) to column table')
def step_impl(context, var_name, list_obj_str):
    var_name = StepTools.evaluate_variable_name(var_name)
    list_obj = StepTools.evaluate_variable_value(list_obj_str)
    
    table = TableManager.convert_object_list_2_column_table(list_obj)
    
    __get_variable_manager().register_variable(var_name, table)

@Then(r"compare table 1 (?P<table_1_str>{Variable}) to table 2 (?P<table_2_str>{Variable}) with(?: \(reorder columns table 1:(?P<reorder_table_1_str>true|false) ; reorder columns table 2:(?P<reorder_table_2_str>true|false)\))?")
def step_impl(context, table_1_str, table_2_str, reorder_table_1_str, reorder_table_2_str):
    table_1 = StepTools.evaluate_variable_value(table_1_str)
    table_2 = StepTools.evaluate_variable_value(table_2_str)
    compare_params_table = BehaveStepTools.convert_step_table_2_value_table_with_header(context.table)
    reorder_columns_table_1 = Converter.to_boolean(reorder_table_1_str.capitalize()) if reorder_table_1_str is not None else True
    reorder_columns_table_2 = Converter.to_boolean(reorder_table_2_str.capitalize()) if reorder_table_2_str is not None else True
    
    TableManager.compare_tables(table_1, table_2, compare_params_table, reorder_columns_table_1, reorder_columns_table_2)

@Step(r"table (?P<table_varname>{Variable}) is")
def step_impl(context, table_varname):
    table_obtained = StepTools.evaluate_variable_value(table_varname)
    BehaveStepTools.then_table_is(table_obtained, context.table)

@Step(r"table (?P<table_varname>{Variable}) is empty")
def step_impl(context, table_varname):
    table = StepTools.evaluate_variable_value(table_varname)
    
    if not table.is_empty:
        raise FunctionalException(f"Table {table_varname} is not empty")

@Step(r"table (?P<table_varname>{Variable}) contains")
def step_impl(context, table_varname):
    table_obtained = StepTools.evaluate_variable_value(table_varname)
    BehaveStepTools.then_table_contains(table_obtained, context.table)

@Step(r"table (?P<table_1_varname>{Variable}) contains table (?P<table_2_varname>{Variable})")
def step_impl(context, table_1_varname, table_2_varname):
    table_1 = StepTools.evaluate_variable_value(table_1_varname)
    table_2 = StepTools.evaluate_variable_value(table_2_varname)
    BehaveStepTools.then_table_contains(table_1, table_2)

@Step(r"table (?P<table_varname>{Variable}) doesn't contain")
def step_impl(context, table_varname):
    table_obtained = StepTools.evaluate_variable_value(table_varname)
    BehaveStepTools.then_table_doesnt_contain(table_obtained, context.table)

@Step(r"table (?P<table_1_varname>{Variable}) doesn't contain table (?P<table_2_varname>{Variable})")
def step_impl(context, table_1_varname, table_2_varname):
    table_1 = StepTools.evaluate_variable_value(table_1_varname)
    table_2 = StepTools.evaluate_variable_value(table_2_varname)
    BehaveStepTools.then_table_doesnt_contain(table_1, table_2)
    
@Given(r'(?P<var_name>{Variable}) = table (?P<table_varname>{Variable}) with columns ordered')
def step_impl(context, var_name, table_varname):
    var_name = StepTools.evaluate_variable_name(var_name)
    table = StepTools.evaluate_variable_value(table_varname)
    table_columns = BehaveStepTools.convert_step_table_2_value_table_with_header(context.table)

    res_table = copy.copy(table)
    columns_ordered = table_columns.get_column_names()
    res_table.order_columns(names=columns_ordered)
    
    __get_variable_manager().register_variable(var_name, res_table)
    
@Given(r"(?P<var_name>{Variable}) = table (?P<table_varname>{Variable}) without columns (?P<col_names>{List})")
def step_impl(context, var_name, table_varname, col_names):
    var_name = StepTools.evaluate_variable_name(var_name)
    table = StepTools.evaluate_variable_value(table_varname)
    col_names = StepTools.evaluate_list_scenario_parameter(col_names, "col_names")
    
    res_table = copy.copy(table)
    for col_name in col_names:
        if res_table.has_column(col_name, raise_exception=False):
            res_table.remove_column(name=col_name)
    
    __get_variable_manager().register_variable(var_name, res_table)
    
@Given(r'(?P<var_name>{Variable}) = table (?P<table_varname>{Variable}) with only columns')
def step_impl(context, var_name, table_varname):
    var_name = StepTools.evaluate_variable_name(var_name)
    table = StepTools.evaluate_variable_value(table_varname)
    table_columns = BehaveStepTools.convert_step_table_2_value_table_with_header(context.table)

    res_table = copy.copy(table)
    expected_colnames = table_columns.get_column_names()
    
    table_colnames = table.get_column_names()
    for col_index, col_name in reversed_enumerate(table_colnames):
        if col_name not in expected_colnames:
            res_table.remove_column(index=col_index)
    
    __get_variable_manager().register_variable(var_name, res_table)
    
@Given(r'(?P<var_name>{Variable}) = table (?P<table_varname>{Variable}) with only columns ordered')
def step_impl(context, var_name, table_varname):
    var_name = StepTools.evaluate_variable_name(var_name)
    rendered_table = render_step_table(context.table, "    ")

    execute_steps(u"""
        Given __TABLE_WITH_ONLY_COLUMNS__@ = table {table_varname} with only columns
        {table}
        """.format(table_varname=table_varname, table=rendered_table))
    
    execute_steps(u"""
        Given {var_name} = table __TABLE_WITH_ONLY_COLUMNS__@ with columns ordered
        {table}
        """.format(var_name=var_name, table=rendered_table))
    
@Given(r'(?P<var_name>{Variable}) = table (?P<table_varname>{Variable}) with new columns')
def step_impl(context, var_name, table_varname):
    var_name = StepTools.evaluate_variable_name(var_name)
    table = StepTools.evaluate_variable_value(table_varname)
    table_new_columns = BehaveStepTools.convert_step_table_2_value_table_with_header(context.table, do_eval_once=False)

    res_table = copy.copy(table)
    is_res_value_table = ValueTableManager.is_value_table(res_table)
    
    res_col_indexes = res_table.get_column_indexes_by_string_content()
    
    tnc_indexes = table_new_columns.get_column_indexes_by_string_content()
    for row_new_column in table_new_columns.rows:
        new_colname_cell = row_new_column[tnc_indexes["Column Name"]]
        new_value_expr_cell = row_new_column[tnc_indexes["Value Expression"]]
        
        # Build new column
        new_col_cells_content = []
        for row in res_table.rows:
            # Add each existing column cells as variable
            for cn in res_col_indexes:
                if is_res_value_table:
                    __get_variable_manager().register_variable(f"Column({cn})", row.get_cell(res_col_indexes[cn]).value, accept_expression=False)
                else:
                    __get_variable_manager().register_variable(f"Column({cn})", row.get_cell(res_col_indexes[cn]).content, accept_expression=False)
                
            # Add new cell
            new_content = new_value_expr_cell.value
            if is_res_value_table and isinstance(new_content, str):
                new_content = f"'{new_content}'"
            new_col_cells_content.append(new_content)
                
        res_table.add_column(name=new_colname_cell.value, cells_content=new_col_cells_content)
    
    __get_variable_manager().register_variable(var_name, res_table)
    
@Given(r'(?P<var_name>{Variable}) = table (?P<table_varname>{Variable}) with new rows')
def step_impl(context, var_name, table_varname):
    var_name = StepTools.evaluate_variable_name(var_name)
    table = StepTools.evaluate_variable_value(table_varname)
    table_new_rows = BehaveStepTools.convert_step_table_2_value_table_with_header(context.table)

    res_table = copy.copy(table)
    is_res_value_table = ValueTableManager.is_value_table(res_table)
    
    # Verify tables headers
    header_comp = StringTableRowComparator()
    header_comp.equals(table_new_rows.header, res_table.header)
    
    # Add rows
    for new_row in table_new_rows.rows:
        cells_value = new_row.cells_value
        if is_res_value_table:
            #TODO: add a test to verify if with this implementation the cells with variables are reevaluated after each compare
            res_table.add_row(cells_value=cells_value)
        else:
            res_table.add_row(cells_content=cells_value)
    
    __get_variable_manager().register_variable(var_name, res_table)
    
@Given(r"(?P<var_name>{Variable}) = table (?P<table_varname>{Variable}) with column (?P<col_name>{Str}) content replaced by")
def step_impl(context, var_name, table_varname, col_name):
    var_name = StepTools.evaluate_variable_name(var_name)
    table = StepTools.evaluate_variable_value(table_varname)
    table_replace = BehaveStepTools.convert_step_table_2_value_table_with_header(context.table, do_eval_once=False)
    col_name = StepTools.evaluate_scenario_parameter(col_name)

    res_table = copy.copy(table)
    is_value_table = ValueTableManager.is_value_table(res_table)
    col = res_table.get_column(name=col_name)
    
    tr_indexes = table_replace.get_column_indexes_by_string_content()
    tcell_comparator = Table2ValueTable_CellComparator() if "Condition Expression" in tr_indexes else None
    for cell in col:
        for row_replace in table_replace.rows:
            cell_cond_value = row_replace[tr_indexes["Condition Value"]] if "Condition Value" in tr_indexes else None
            cell_cond_expr = row_replace[tr_indexes["Condition Expression"]] if "Condition Expression" in tr_indexes else None
            cell_rep_value = row_replace[tr_indexes["Replace Value"]] if "Replace Value" in tr_indexes else None
            
            if cell_cond_value is not None:
                # If condition value is 'DEFAULT' or cell value is equal to condition value
                cond = (cell_cond_value.content_type == ValueTypes.Symbol and cell_cond_value.content == 'DEFAULT'
                        or cell_cond_value.value_type != ValueTypes.NotApplicable 
                            and (is_value_table and cell.value == cell_cond_value.value or not is_value_table and cell.content == cell_cond_value.value) )
            elif cell_cond_expr is not None:
                cond = tcell_comparator.equals(cell, cell_cond_expr, raise_exception=False, do_log=False)
            else:
                continue
                
            if cond:
                if cell_rep_value.value_type != ValueTypes.NotApplicable:
                    # Note: As 'col' is created with a copy of 'res_table' cells, 'cell' is also a 'res_table' cell.
                    #       Thus, modifying 'cell' is modifying the cell in 'res_table'
                    if is_value_table:
                        cell.value = cell_rep_value.value
                    else:
                        cell.content = cell_rep_value.value
                    break
                else:
                    continue
                
    __get_variable_manager().register_variable(var_name, res_table)
    
@Given(r"(?P<var_name>{Variable}) = table (?P<table_varname>{Variable}) with columns (?P<col_names>{List}) content replaced by")
def step_impl(context, var_name, table_varname, col_names):
    var_name = StepTools.evaluate_variable_name(var_name)
    table = StepTools.evaluate_variable_value(table_varname)
    col_names = StepTools.evaluate_list_scenario_parameter(col_names, "col_names")
    
    rendered_table = render_step_table(context.table, "    ")

    __get_variable_manager().register_variable(var_name, copy.copy(table))
    
    for col_name in col_names:
        execute_steps(u"""
            Given {var_name} = table {var_name} with column '{col_name}' content replaced by
            {table}
            """.format(var_name=var_name, col_name=col_name, table=rendered_table))
    
@Given(r"(?P<var_name>{Variable}) = table (?P<table_varname>{Variable}) with rows ordered by (?P<col_names>{List})")
def step_impl(context, var_name, table_varname, col_names):
    var_name = StepTools.evaluate_variable_name(var_name)
    table = StepTools.evaluate_variable_value(table_varname)
    col_names = StepTools.evaluate_list_scenario_parameter(col_names, "col_names")

    res_table = copy.copy(table)
    res_table.sort(names=col_names)
    
    __get_variable_manager().register_variable(var_name, res_table)
    
@Given(r"(?P<var_name>{Variable}) = table (?P<table_varname>{Variable}) with rows verifying")
def step_impl(context, var_name, table_varname):
    """
    Keep only rows verifying at least one line of step table.
    
    If table table_varname is without header, step table can't have a header and must have same column number as 'table_varname'. 
    If table table_varname is with header, step table must be a table with header, but it can contain only some columns of 'table_varname'. 
    """
    var_name = StepTools.evaluate_variable_name(var_name)
    table = StepTools.evaluate_variable_value(table_varname)
    is_with_header = TableManager.is_table_with_header(table)
    if is_with_header:
        table_remove = BehaveStepTools.convert_step_table_2_value_table_with_header(context.table)
    else:
        table_remove = BehaveStepTools.convert_step_table_2_value_table(context.table)
    
    if is_with_header:
        comparator = Table2ValueTable_WithHeaderComparator()
    else:
        comparator = Table2ValueTable_Comparator()

    res_table = copy.copy(table)
    if is_with_header and table_remove.nb_columns != table.nb_columns:
        res_table.remove_rows_only_verifying(table_remove, comparator, keep_rows=True)
    else:
        res_table.remove_rows_verifying(table_remove, comparator, keep_rows=True)
    
    __get_variable_manager().register_variable(var_name, res_table)
    
@Given(r"(?P<var_name>{Variable}) = table (?P<table_varname>{Variable}) without rows verifying")
def step_impl(context, var_name, table_varname):
    """
    Remove rows verifying at least one line of step table.
    
    If table table_varname is without header, step table can't have a header and must have same column number as 'table_varname'. 
    If table table_varname is with header, step table must be a table with header, but it can contain only some columns of 'table_varname'. 
    """
    var_name = StepTools.evaluate_variable_name(var_name)
    table = StepTools.evaluate_variable_value(table_varname)
    is_with_header = TableManager.is_table_with_header(table)
    if is_with_header:
        table_remove = BehaveStepTools.convert_step_table_2_value_table_with_header(context.table)
    else:
        table_remove = BehaveStepTools.convert_step_table_2_value_table(context.table)
    
    if is_with_header:
        comparator = Table2ValueTable_WithHeaderComparator()
    else:
        comparator = Table2ValueTable_Comparator()

    res_table = copy.copy(table)
    if is_with_header and table_remove.nb_columns != table.nb_columns:
        res_table.remove_rows_only_verifying(table_remove, comparator)
    else:
        res_table.remove_rows_verifying(table_remove, comparator)
    
    __get_variable_manager().register_variable(var_name, res_table)
    
@Given(r"(?P<var_name>{Variable}) = table (?P<table_varname>{Variable}) without duplicated rows")
def step_impl(context, var_name, table_varname):
    """
    Remove duplicated rows considering only column names in step table.
    """
    var_name = StepTools.evaluate_variable_name(var_name)
    table = StepTools.evaluate_variable_value(table_varname)
    table_columns = BehaveStepTools.convert_step_table_2_value_table_with_header(context.table)

    res_table = copy.copy(table)
    res_table.remove_rows_duplicated(column_names=table_columns.get_column_names())
    
    __get_variable_manager().register_variable(var_name, res_table)
    
@Given(r"(?P<var_name>{Variable}) = extract column (?P<col_name>{Str}) from table (?P<table_varname>{Variable})")
def step_impl(context, var_name, table_varname, col_name):
    var_name = StepTools.evaluate_variable_name(var_name)
    col_name = StepTools.evaluate_scenario_parameter(col_name)
    table = StepTools.evaluate_variable_value(table_varname)

    if isinstance(table, TableWithHeader):
        res_table = TableWithHeader()
        res = table.get_column(name=col_name)

        res_table.header = TableRow(cells_content={col_name})
        for c in res.cells:
            res_table.add_row(cells_content={c.content})
       
        __get_variable_manager().register_variable(var_name, res_table)
    else:
        raise TechnicalException(f"Table {table_varname} as no column {col_name}")
    
#TODO: return a list rather than a table
@Given(r"(?P<var_name>{Variable}) = extract column (?P<col_name>{Str}) cells from table (?P<table_varname>{Variable})")
def step_impl(context, var_name, table_varname, col_name):
    var_name = StepTools.evaluate_variable_name(var_name)
    col_name = StepTools.evaluate_scenario_parameter(col_name)
    table = StepTools.evaluate_variable_value(table_varname)

    if isinstance(table, TableWithHeader):
        res_table = Table()
        res = table.get_column(name=col_name)

        for c in res.cells:
            res_table.add_row(cells_content={c.content})
       
        __get_variable_manager().register_variable(var_name, res_table)
    else:
        raise TechnicalException(f"Table {table_varname} as no column {col_name}")
    
#TODO: remove
@Given(r"(?P<var_name>{Variable}) = extract column (?P<col_name>{Str}) cells from table (?P<table_varname>{Variable}) as row")
def step_impl(context, var_name, table_varname, col_name):
    var_name = StepTools.evaluate_variable_name(var_name)
    col_name = StepTools.evaluate_scenario_parameter(col_name)
    table = StepTools.evaluate_variable_value(table_varname)

    if isinstance(table, TableWithHeader):
        res_table = Table()
        res = table.get_column(name=col_name)

        res_table.add_row()
        for c in res.cells:
            res_table.add_column(cells_content={c.content})
       
        __get_variable_manager().register_variable(var_name, res_table)
    else:
        raise TechnicalException(f"Table {table_varname} as no column {col_name}")
        

@Given(r"(?P<var_name>{Variable}) = table (?P<table_varname>{Variable}) with columns renamed")
def step_impl(context, var_name, table_varname):
    var_name = StepTools.evaluate_variable_name(var_name)
    table = StepTools.evaluate_variable_value(table_varname)
    table_columns = BehaveStepTools.convert_step_table_2_value_table(context.table)
    
    col_index = table.get_column_indexes_by_string_content()
    
    res_table = copy.copy(table)
    header = res_table.header

    for row in table_columns.rows:
        old_name = row.get_cell(0).string_content
        new_name = row.get_cell(1).string_content
        
        if old_name in col_index:
            header.get_cell(col_index[old_name]).content = new_name
    res_table.header = header
    
    __get_variable_manager().register_variable(var_name, res_table)
    
@Given(r"(?P<var_name>{Variable}) = number of rows in table (?P<table_varname>{Variable})")
def step_impl(context, var_name, table_varname):
    var_name = StepTools.evaluate_variable_name(var_name)
    table = StepTools.evaluate_variable_value(table_varname)
    
    # print(table.represent())
    count_table = table.nb_rows
    
    # print(count_table)    
    __get_variable_manager().register_variable(var_name, count_table)
    
    
    


# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

import account_pb2 as account__pb2
from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2


class UserControllerStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.List = channel.unary_stream(
                '/account.UserController/List',
                request_serializer=account__pb2.UserListRequest.SerializeToString,
                response_deserializer=account__pb2.User.FromString,
                )
        self.Create = channel.unary_unary(
                '/account.UserController/Create',
                request_serializer=account__pb2.User.SerializeToString,
                response_deserializer=account__pb2.User.FromString,
                )
        self.Retrieve = channel.unary_unary(
                '/account.UserController/Retrieve',
                request_serializer=account__pb2.UserRetrieveRequest.SerializeToString,
                response_deserializer=account__pb2.User.FromString,
                )
        self.Update = channel.unary_unary(
                '/account.UserController/Update',
                request_serializer=account__pb2.User.SerializeToString,
                response_deserializer=account__pb2.User.FromString,
                )
        self.Destroy = channel.unary_unary(
                '/account.UserController/Destroy',
                request_serializer=account__pb2.User.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                )


class UserControllerServicer(object):
    """Missing associated documentation comment in .proto file."""

    def List(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Create(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Retrieve(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Update(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Destroy(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_UserControllerServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'List': grpc.unary_stream_rpc_method_handler(
                    servicer.List,
                    request_deserializer=account__pb2.UserListRequest.FromString,
                    response_serializer=account__pb2.User.SerializeToString,
            ),
            'Create': grpc.unary_unary_rpc_method_handler(
                    servicer.Create,
                    request_deserializer=account__pb2.User.FromString,
                    response_serializer=account__pb2.User.SerializeToString,
            ),
            'Retrieve': grpc.unary_unary_rpc_method_handler(
                    servicer.Retrieve,
                    request_deserializer=account__pb2.UserRetrieveRequest.FromString,
                    response_serializer=account__pb2.User.SerializeToString,
            ),
            'Update': grpc.unary_unary_rpc_method_handler(
                    servicer.Update,
                    request_deserializer=account__pb2.User.FromString,
                    response_serializer=account__pb2.User.SerializeToString,
            ),
            'Destroy': grpc.unary_unary_rpc_method_handler(
                    servicer.Destroy,
                    request_deserializer=account__pb2.User.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'account.UserController', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class UserController(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def List(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/account.UserController/List',
            account__pb2.UserListRequest.SerializeToString,
            account__pb2.User.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Create(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/account.UserController/Create',
            account__pb2.User.SerializeToString,
            account__pb2.User.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Retrieve(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/account.UserController/Retrieve',
            account__pb2.UserRetrieveRequest.SerializeToString,
            account__pb2.User.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Update(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/account.UserController/Update',
            account__pb2.User.SerializeToString,
            account__pb2.User.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Destroy(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/account.UserController/Destroy',
            account__pb2.User.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)




#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
import os
from holado_core.common.exceptions.functional_exception import FunctionalException
from holado_core.common.exceptions.technical_exception import TechnicalException
import sys
import importlib
from holado_core.common.tools.tools import Tools

logger = logging.getLogger(__name__)

try:
    import grpc_requests  # @UnusedImport
    with_grpc_requests = True
except Exception as exc:
    if Tools.do_log(logger, logging.DEBUG):
        logger.debug(f"GRpcServices is not available. Initialization failed on error: {exc}")
    with_grpc_requests = False


class GRpcServices(object):
    @classmethod
    def is_available(cls):
        return with_grpc_requests
    
    def __init__(self): 
        self.__service_descriptors_by_fullname = {}
        self.__services_by_fullname = {}
    
    def import_all_compiled_proto(self, compiled_proto_path, raise_if_not_exist=True):
        """Register a folder path containing compiled proto files. Usually it corresponds to the parameter '--python_out' passed to proto compiler."""
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"[GRpcServices] Importing all compiled proto in '{compiled_proto_path}'...")
        if os.path.exists(compiled_proto_path):
            if os.path.isfile(compiled_proto_path):
                proto_path = os.path.dirname(compiled_proto_path)
                sys.path.append(proto_path)
                self.__import_compiled_proto(proto_path, compiled_proto_path)
            elif os.path.isdir(compiled_proto_path):
                sys.path.append(compiled_proto_path)
                self.__import_all_compiled_proto(compiled_proto_path, "")
            else:
                raise TechnicalException(f"Unmanaged path '{proto_path}'")
        else:
            msg = f"Path '{compiled_proto_path}' doesn't exist"
            if raise_if_not_exist:
                raise TechnicalException(msg)
            else:
                logger.warning(msg)
    
    def __import_all_compiled_proto(self, compiled_proto_path, package_name):
        if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
            logger.trace(f"[GRpcServices] Importing all compiled proto in '{compiled_proto_path}' (package: '{package_name}')...")
        if os.path.isdir(compiled_proto_path):
            lp = os.listdir(compiled_proto_path)
            for cp in lp:
                if not cp.startswith((".", "_")):
                    cur_proto_path = os.path.join(compiled_proto_path, cp)
                    
                    if os.path.isfile(cur_proto_path):
                        self.__import_compiled_proto(cur_proto_path, package_name)
                    elif os.path.isdir(cur_proto_path):
                        cur_package_name = f"{package_name}.{cp}" if package_name is not None and len(package_name) > 0 else cp
                        self.__import_all_compiled_proto(cur_proto_path, cur_package_name)
                    else:
                        raise TechnicalException(f"Unmanaged path '{cur_proto_path}'")
        else:
            raise TechnicalException(f"Unmanaged path '{compiled_proto_path}'")
    
    def __import_compiled_proto(self, compiled_proto_file_path, package_name):
        if not os.path.isfile(compiled_proto_file_path):
            raise TechnicalException(f"Compiled proto path '{compiled_proto_file_path}' is not a file")
        if not compiled_proto_file_path.endswith("_pb2.py"):
            return

        if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
            logger.trace(f"[GRpcServices] Importing compiled proto file '{compiled_proto_file_path}' (package: '{package_name}')...")
        
        filename = os.path.splitext(os.path.basename(compiled_proto_file_path))[0]
        module_name = f"{package_name}.{filename}" if package_name is not None and len(package_name) > 0 else filename
        if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
            logger.trace(f"[GRpcServices] Importing module '{module_name}'")
        module = importlib.import_module(module_name)
        
        if hasattr(module.DESCRIPTOR, 'services_by_name'):
            service_package = module.DESCRIPTOR.package if hasattr(module.DESCRIPTOR, 'package') else package_name
            
            module_grpc_name = module_name + '_grpc'
            module_spec = importlib.util.find_spec(module_grpc_name)
            if module_spec is not None:
                module_grpc = importlib.import_module(module_grpc_name)
            
            for s_name in module.DESCRIPTOR.services_by_name:
                s_fullname = f"{service_package}.{s_name}" if service_package is not None and len(service_package) > 0 else s_name
                self.__service_descriptors_by_fullname[s_fullname] = module.DESCRIPTOR.services_by_name[s_name]
                if Tools.do_log(logger, logging.DEBUG):
                    logger.debug(f"[GRpcServices] New managed service descriptor '{s_fullname}'")
                if module_grpc is not None:
                    self.__services_by_fullname[s_fullname] = getattr(module_grpc, s_name)
                    if Tools.do_log(logger, logging.DEBUG):
                        logger.debug(f"[GRpcServices] New managed service '{s_fullname}' (type: '{self.__services_by_fullname[s_fullname].__qualname__}')")

    def has_service(self, service_fullname):
        """Return if service fullname is known."""
        return service_fullname in self.__services_by_fullname

    def get_service(self, service_fullname):
        """Return service for given service fullname."""
        if self.has_service(service_fullname):
            return self.__services_by_fullname[service_fullname]
        else:
            raise FunctionalException(f"Unknown service '{service_fullname}' (existing services: {list(self.__services_by_fullname.keys())})")

    def has_service_descriptor(self, service_fullname):
        """Return if service descriptor fullname is known."""
        return service_fullname in self.__service_descriptors_by_fullname

    def get_service_descriptor(self, service_fullname):
        """Return service descriptor for given service fullname."""
        if self.has_service_descriptor(service_fullname):
            return self.__service_descriptors_by_fullname[service_fullname]
        else:
            raise FunctionalException(f"Unknown service '{service_fullname}' (existing services: {list(self.__service_descriptors_by_fullname.keys())})")

    def has_method_descriptor(self, service_fullname, method_name):
        """Return if method exists in service descriptor fullname."""
        service_descriptor = self.get_service_descriptor(service_fullname)
        return method_name in service_descriptor.methods_by_name

    def get_method_descriptor(self, service_fullname, method_name):
        """Return service method descriptor for given service fullname and method name."""
        # Get service descriptor to verify in same type that it exists
        service_descriptor = self.get_service_descriptor(service_fullname)
        
        # Get method descriptor
        if method_name in service_descriptor.methods_by_name:
            return service_descriptor.methods_by_name[method_name]
        else:
            raise FunctionalException(f"Unknown method '{method_name}' in service '{service_fullname}' (existing methods in service: {list(service_descriptor.methods_by_name.keys())})")
    
    def get_all_service_fullnames(self):
        return tuple(self.__service_descriptors_by_fullname.keys())
    
    def get_all_method_fullnames(self):
        res = []
        for service_fullname in self.get_all_service_fullnames():
            service_descriptor = self.get_service_descriptor(service_fullname)
            for method_name in service_descriptor.methods_by_name:
                res.append(f"{service_fullname}.{method_name}")
        return res
    
    
        


#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from holado_core.common.inspectors.tree_inspector import TreeInspector
from holado_core.common.inspectors.tools.inspect_parameters import InspectParameters
from holado_ui.ui.handlers.ui_context import UIContext
from holado_ui.ui.inspectors.tools.ui_inspect_context import UIInspectContext
import abc

logger = logging.getLogger(__name__)



class UIInspector(TreeInspector):
    """ Base class for inspector on UI elements.
    """
    __metaclass__ = abc.ABCMeta
    
    def __init__(self, module_name):
        super().__init__(module_name)
        
        self.__driver = None
    
    def initialize(self, driver):
        self.__driver = driver
        
    @property
    def driver(self):
        return self.__driver
        
    @property
    def internal_driver(self):
        return self.driver.internal_driver
    
    def _build_result_finder(self, finder, inspect_context, inspect_parameters):
        ui_context = inspect_context.ui_context()
        if ui_context is not None:
            return self.get_finder_element_in_ui_context(ui_context, finder, inspect_context.without_ui_context(), inspect_parameters)
        else:
            return super()._build_result_finder(finder, inspect_context, inspect_parameters)

    def get_finder_ui_context(self, ui_context:UIContext, inspect_context:UIInspectContext=None, inspect_parameters:InspectParameters=None):
        """
        Finder of UI context.
        @param ui_context Context
        @param inspect_context Inspect context
        @param inspect_parameters Inspect parameters
        @return Finder
        """
        if ui_context is None:
            return self._build_result_finder(None, inspect_context, inspect_parameters)
        else:
            return self.get_finder_element_in_ui_context(ui_context.get_previous_context(),
                self._get_finder_ui_context_element(ui_context, inspect_context.without_ui_context(), inspect_parameters), 
                inspect_context, inspect_parameters)
    
    def _get_finder_ui_context_element(self, ui_context:UIContext, inspect_context:UIInspectContext=None, inspect_parameters:InspectParameters=None):
        """
        
        @param context UI context
        @param inspect_context Inspect context
        @param inspect_parameters Inspect parameters
        @return Finder of given UI context, without consideration of previous UI context
        """
        raise NotImplementedError

    def get_finder_element_in_ui_context(self, ui_context:UIContext, finder_element, inspect_context:UIInspectContext=None, inspect_parameters:InspectParameters=None):
        """
        Find element with given finder, in given context
        @param ui_context Context
        @param finder_element Finder for element to find
        @param inspect_context Inspect context
        @param inspect_parameters Inspect parameters
        @return Finder
        """
        return self.get_finder_element_in(
            self.get_finder_ui_context(ui_context, inspect_context.without_ui_context(), inspect_parameters), 
            finder_element, inspect_context, inspect_parameters)
    
    
    
    